# lsqfitgp/_patch_autograd.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""Add vjp/jvp of asarray, fix #552 (creating object arrays from lists),
add missing scipy functions, add numpy.testing"""

import builtins

import autograd
import numpy as np

autograd.extend.defvjp(
    autograd.numpy.asarray,
    lambda ans, *args, **kw: lambda g: g
)
autograd.extend.defjvp(
    autograd.numpy.asarray,
    lambda g, ans, *args, **kw: g
)

autograd.numpy.numpy_boxes.ArrayBox.item = lambda self: self[(0,) * len(self.shape)]

def array(A, *args, **kwargs):
    t = builtins.type(A)
    if t in (list, tuple):
        return autograd.numpy.numpy_wrapper.array_from_args(args, kwargs, *map(lambda a: a if a.shape else a.item(), map(array, A)))
    else:
        return autograd.numpy.numpy_wrapper._array_from_scalar_or_array(args, kwargs, A)
autograd.numpy.numpy_wrapper.array = array
autograd.numpy.array = array

autograd.numpy.testing = np.testing

from scipy import special as special_noderiv
from autograd.scipy import special

missing_functions = [
    'bernoulli',
    'binom',
    'kv',
    'kvp',
    'factorial',
    'zeta'
]
for f in missing_functions:
    if not hasattr(special, f):
        wf = autograd.extend.primitive(getattr(special_noderiv, f))
        setattr(special, f, wf)

autograd.extend.defvjp(
    special.kvp,
    lambda ans, v, z, n: lambda g: g * special.kvp(v, z, n + 1),
    argnums=[1]
)    



# lsqfitgp/__init__.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from . import _patch_autograd
from . import _gvar_autograd

from autograd import numpy, scipy

from ._GP import *
from ._Kernel import *
from ._kernels import *
from ._array import *
from ._fit import *
from ._Deriv import *
from ._fastraniter import *

__version__ = '0.8'

__doc__ = """

Module to fit Gaussian processes with gvar/lsqfit.

Manual: https://gattocrucco.github.io/lsqfitgp/docs

It can both be used standalone to fit data with a Gaussian process only, and
with lsqfit inside a possibly nonlinear model with other parameters. In lsqfit
style, all the results will be properly correlated with prior, data, and other
non-Gaussian process parameters in the fit, even when doing conditional
prediction.

The main class is `GP`, which represents a Gaussian process over arbitrary
input. It can be used both autonomously and with lsqfit. The inputs/outputs can
be arrays or dictionaries of arrays. It supports doing inference with the
derivatives of the process, using `autograd` to compute automatically
derivatives of the kernels. Indirectly, this can be used to make inference with
integrals.

Functions and classes
---------------------
    
    GP : class
        Class of objects representing a Gaussian process.
    empbayes_fit : class
        Fit the hyperparameters of a Gaussian process.
    StructuredArray : class
        Autograd-friendly wrapper of numpy structured arrays.
    where : function
        Make a kernel that switches between two kernels based on a condition.
    Deriv : class
        Class representing a derivative specification.

Kernels
-------

The covariance functions are represented by subclasses of class `Kernel`. There
are also `IsotropicKernel` for covariance functions that depend only on the
distance between the arguments, and `StationaryKernel` for kernels that depend
only on the difference. Kernel objects can be summed, multiplied and raised to
a power.

To make a custom kernel, you can instantiate one of the three general classes
by passing them a function, or subclass them. For convenience, decorators
`kernel`, `stationarykernel` and `isotropickernel` are provided to convert a
function to a covariance function. Otherwise, use one of the already available
subclasses listed below. Isotropic and stationary kernels are normalized to
have unit variance and roughly unit lengthscale.

    Constant
        Equivalent to fitting with a constant.
    Linear
        Equivalent to fitting with a line.
    ExpQuad
        Gaussian kernel.
    White
        White noise, each point is indipendent.
    Matern
        Matérn kernel, you can set how many times it is differentiable.
    Matern12, Matern32, Matern52
        Matérn kernel for the specific cases nu = 1/2, 3/2, 5/2.
    GammaExp
        Gamma exponential. Not differentiable, but you can set how close it is
        to being differentiable.
    RatQuad
        Equivalent to a mixture of Gaussian kernels with gamma-distributed
        length scales.
    NNKernel
        Equivalent to training a neural network with one latent infinite layer.
    Wiener
        Random walk.
    WienerIntegral
        Integral of the random walk.
    FracBrownian
        Fractional Brownian motion, like Wiener but with correlations.
    OrnsteinUhlenbeck
        Random walk with asymptotically finite variance.
    BrownianBridge
        Random walk which comes back to the starting point.
    Gibbs
        A Gaussian kernel with a custom variable length scale.
    Categorical
        Arbitrary covariance matrix over a finite set of values.
    Cos
        A cosine.
    Celerite
        A pair of conjugated complex exponentials.
    Harmonic
        Stochastically driven damped harmonic oscillator.
    PPKernel
        Finite support isotropic kernel.
    Rescaling
        Kernel used to change the variance of other kernels.
    Taylor
        Exponential-like taylor series.
    Fourier
        Kernel for periodic functions, the decay of the Fourier coefficients is
        adjustable.
    Periodic
        A periodic Gaussian kernel, represents a periodic function.

Reference: Rasmussen et al. (2006), "Gaussian Processes for Machine Learning".

"""



# lsqfitgp/_toeplitz_linalg.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# This code originally copied from the TOEPLITZ_CHOLESKY code by John Burkardt,
# released under the LGPL.
# people.sc.fsu.edu/~jburkardt/py_src/toeplitz_cholesky/toeplitz_cholesky.html

import numpy as np

def cholesky(t, diageps=None):
    """
    
    Cholesky decomposition of a symmetric Toeplitz matrix.
    
    Parameters
    ----------
    t : (N,) array
        The first row/column of the matrix.
    diageps: float, optional
        The diagonal of the matrix (the first element of t) is increased by
        `diageps`.
    
    Returns
    -------
    L : (N, N) array
        Lower triangular matrix such that toeplitz(t) == L @ L.T.
    
    Notes
    -----
    The reference for the algorithm is:
        
        Michael Stewart,
        Cholesky factorization of semi-definite Toeplitz matrices.
        Linear Algebra and its Applications,
        Volume 254, pages 497-525, 1997.
    
    """
    
    # TODO Split the function in an interface and a numba core. Exceptions are
    # raised by the interface, the core returns the last i.
    
    # TODO Make a function to compute L @ M instead of just computing L, and
    # one to do L^-1 @ M. Both matrix multiplication and triangular solve can
    # be done one column at a time.
    
    # TODO Probably if I compile this with numba the code as originally
    # written with explicit indices is faster.
    
    t = np.asanyarray(t)
    assert len(t.shape) == 1
    if len(t) == 0:
        return np.empty((0, 0), dtype=t.dtype)
    if t[0] <= 0:
        msg = '1-th leading minor is not positive definite'
        raise np.linalg.LinAlgError(msg)

    n = len(t)
    L = np.zeros((n, n), order='F', dtype=t.dtype)
    L[:, 0] = t
    if diageps is not None:
        t = L[:, 0]
        t[0] += diageps
    g = np.stack([np.roll(t, 1), t])

    for i in range(1, n):
        assert g[0, i] > 0
        rho = -g[1, i] / g[0, i]
        if np.abs(rho) >= 1:
            msg = '{}-th leading minor is not positive definite'.format(i + 1)
            raise np.linalg.LinAlgError(msg)
        gamma = np.sqrt((1 - rho) * (1 + rho))
        g[:, i:] += g[::-1, i:] * rho
        g[:, i:] /= gamma
        L[i:, i] = g[0, i:]
        g[0, i:] = np.roll(g[0, i:], 1)

    return L

def eigv_bound(t):
    """
    
    Bound the eigenvalues of a symmetric Toeplitz matrix.
    
    Parameters
    ----------
    t : array
        The first row of the matrix.
    
    Returns
    -------
    m : scalar
        Any eigenvalue `v` of the matrix satisfies `|v| <= m`.
    
    """
    s = np.abs(t)
    c = np.cumsum(s)
    d = c + c[::-1] - s[0]
    return np.max(d)



# lsqfitgp/_kernels.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import re
import collections

import autograd
from autograd import numpy as np
from autograd.scipy import special
import numpy # to bypass autograd

from . import _array
from . import _Kernel
from ._Kernel import kernel, stationarykernel, isotropickernel

__all__ = [
    'Constant',
    'White',
    'ExpQuad',
    'Linear',
    'Matern',
    'Matern12',
    'Matern32',
    'Matern52',
    'GammaExp',
    'RatQuad',
    'NNKernel',
    'Wiener',
    'Gibbs',
    'Periodic',
    'Categorical',
    'Rescaling',
    'Cos',
    'FracBrownian',
    'PPKernel',
    'WienerIntegral',
    'Taylor',
    'Fourier',
    'OrnsteinUhlenbeck',
    'Celerite',
    'BrownianBridge',
    'Harmonic',
    'Expon',
    'BagOfWords',
]

def _dot(x, y):
    return _Kernel._sum_recurse_dtype(lambda x, y: x * y, x, y)
    
@isotropickernel(derivable=True)
def Constant(r2):
    """
    Constant kernel.
    
    .. math::
        k(r) = 1
    
    This means that all points are completely correlated, thus it is equivalent
    to fitting with an horizontal line. This can be seen also by observing that
    1 = 1 x 1.
    """
    return np.ones_like(r2)
    # TODO maybe use dtype=int8 as optimization? => not worth it
    
@isotropickernel(derivable=False)
def White(r2):
    """
    White noise kernel.
    
    .. math::
        k(x, y) = \\begin{cases}
            1 & x = y     \\\\
            0 & x \\neq y
        \\end{cases}
    """
    # TODO add support for input that doesn't have a distance but
    # just equality, like for example non-numerical input.
    return np.where(r2 == 0, 1, 0)

@isotropickernel(derivable=True)
def ExpQuad(r2):
    """
    Exponential quadratic kernel.
    
    .. math::
        k(r) = \\exp \\left( -\\frac 12 r^2 \\right)
    
    It is smooth and has a strict typical lengthscale, i.e., oscillations are
    strongly suppressed under a certain wavelength, and correlations are
    strongly suppressed over a certain distance.
    """
    return np.exp(-1/2 * r2)

@kernel(derivable=True)
def Linear(x, y):
    """
    Dot product kernel.
    
    .. math::
        k(x, y) = x \\cdot y = \\sum_i x_i y_i
    
    In 1D it is equivalent to fitting with a line passing by the origin.
    """
    return _dot(x, y)

@autograd.extend.primitive
def _maternp(x, p):
    poly = 1
    for k in reversed(range(p)):
        c_kp1_over_ck = (p - k) / ((2 * p - k) * (k + 1))
        poly *= c_kp1_over_ck * 2 * x
        poly += 1
    return np.exp(-x) * poly

def _maternp_deriv(x, p):
    if p == 0:
        return -np.exp(-x)
    poly = 1
    for k in reversed(range(1, p)):
        c_kp1_over_ck = (p - k) / ((2 * p - k - 1) * k)
        poly = 1 + poly * c_kp1_over_ck * 2 * x
    poly = poly / (1 - 2 * p) * x
    return np.exp(-x) * poly

autograd.extend.defvjp(
    _maternp,
    lambda ans, x, p: lambda g: g * _maternp_deriv(x, p)
)

def _matern_derivable(**kw):
    nu = kw.get('nu', None)
    if np.isscalar(nu) and nu > 0 and (2 * nu) % 1 == 0:
        return int(nu - 1/2)
    else:
        return False

# TODO I'm using soft input for the Matérn kernels, however for the
# half-integer case it is probably not necessary to use the softabs. Add an
# option 'hard' to the IsotropicKernel init and see if it works.

@isotropickernel(input='soft', derivable=_matern_derivable)
def Matern(r, nu=None):
    """
    Matérn kernel of real order. 
    
    .. math::
        k(r) = \\frac {2^{1-\\nu}} {\\Gamma(\\nu)} x^\\nu K_\\nu(x),
        \\quad \\nu = \\texttt{nu} > 0,
        \\quad x = \\sqrt{2\\nu} r
    
    The nearest integer below `nu` indicates how many times the Gaussian
    process is derivable: so for `nu` < 1 it is continuous but not derivable,
    for 1 <= `nu` < 2 it is derivable but has not a decond derivative, etc. The
    half-integer case (nu = 1/2, 3/2, ...) uses internally a simpler formula so
    you should prefer it. Also, taking derivatives of the process is supported
    only for half-integer nu.

    """
    assert np.isscalar(nu)
    assert nu > 0
    x = np.sqrt(2 * nu) * r
    if (2 * nu) % 1 == 0:
        return _maternp(x, int(nu - 1/2))
    else:
        return 2 ** (1 - nu) / special.gamma(nu) * x ** nu * special.kv(nu, x)

@isotropickernel(input='soft', derivable=False)
def Matern12(r):
    """
    Matérn kernel of order 1/2 (continuous, not derivable).
    
    .. math::
        k(r) = \\exp(-r)
    """
    return np.exp(-r)

@autograd.extend.primitive
def _matern32(x):
    return (1 + x) * np.exp(-x)

autograd.extend.defvjp(
    _matern32,
    lambda ans, x: lambda g: g * -x * np.exp(-x)
)

autograd.extend.defjvp(
    _matern32,
    lambda g, ans, x: (g.T * (-x * np.exp(-x)).T).T
)

@isotropickernel(input='soft', derivable=1)
def Matern32(r):
    """
    Matérn kernel of order 3/2 (derivable one time).
    
    .. math::
        k(r) = (1 + x) \\exp(-x), \\quad x = \\sqrt3 r
    """
    return _matern32(np.sqrt(3) * r)

@autograd.extend.primitive
def _matern52(x):
    return (1 + x * (1 + x/3)) * np.exp(-x)

autograd.extend.defvjp(
    _matern52,
    lambda ans, x: lambda g: g * -x/3 * _matern32(x)
)

@isotropickernel(input='soft', derivable=2)
def Matern52(r):
    """
    Matérn kernel of order 5/2 (derivable two times).
    
    .. math::
        k(r) = (1 + x + x^2/3) \\exp(-x), \\quad x = \\sqrt5 r
    """
    return _matern52(np.sqrt(5) * r)
    
def _gammaexp_derivable(gamma=1):
    return gamma == 2

@isotropickernel(input='soft', derivable=_gammaexp_derivable)
def GammaExp(r, gamma=1):
    """
    Gamma exponential kernel.
    
    .. math::
        k(r) = \\exp(-r^\\texttt{gamma}), \\quad
        \\texttt{gamma} \\in [0, 2]
    
    For `gamma` = 2 it is the Gaussian kernel, for `gamma` = 1 it is the Matérn
    1/2 kernel, for `gamma` = 0 it is the constant kernel. The process is
    differentiable only for `gamma` = 2, however as `gamma` gets closer to 2
    the variance of the non-derivable component goes to zero.

    """
    assert np.isscalar(gamma)
    assert 0 <= gamma <= 2
    return np.exp(-(r ** gamma))

@isotropickernel(derivable=True)
def RatQuad(r2, alpha=2):
    """
    Rational quadratic kernel.
    
    .. math::
        k(r) = \\left( 1 + \\frac {r^2} {2 \\alpha} \\right)^{-\\alpha},
        \\quad \\alpha = \\texttt{alpha}
    
    It is equivalent to a lengthscale mixture of Gaussian kernels where the
    scale distribution is a gamma with shape parameter `alpha`. For `alpha` ->
    infinity, it becomes the Gaussian kernel. It is smooth.
    
    """
    assert np.isscalar(alpha)
    assert 0 < alpha < np.inf
    return (1 + r2 / (2 * alpha)) ** -alpha

@kernel(derivable=True)
def NNKernel(x, y, sigma0=1):
    """
    Neural network kernel.
    
    .. math::
        k(x, y) = \\frac 2 \\pi
        \\arcsin \\left( \\frac
        {
            2 (q + x \\cdot y)
        }{
            (1 + 2 (q + x \\cdot x))
            (1 + 2 (q + y \\cdot y))
        }
        \\right),
        \\quad q = \\texttt{sigma0}^2
    
    Kernel which is equivalent to a neural network with one infinite hidden
    layer with Gaussian priors on the weights and error function response. In
    other words, you can think of the process as a superposition of sigmoids
    where `sigma0` sets the dispersion of the centers of the sigmoids.
    
    """
    
    # TODO the `2`s in the formula are a bit arbitrary. Remove them or give
    # motivation relative to the precise formulation of the neural network.
    
    assert np.isscalar(sigma0)
    assert np.isfinite(sigma0)
    assert sigma0 > 0
    q = sigma0 ** 2
    denom = (1 + 2 * (q + _dot(x, x))) * (1 + 2 * (q + _dot(y, y)))
    return 2/np.pi * np.arcsin(2 * (q + _dot(x, y)) / denom)
    
    # TODO this is not fully equivalent to an arbitrary transformation on the
    # augmented vector even if x and y are transformed, unless I support q
    # being a vector or an additional parameter.

@kernel(forcekron=True, derivable=False)
def Wiener(x, y):
    """
    Wiener kernel.
    
    .. math::
        k(x, y) = \\min(x, y), \\quad x, y > 0
    
    A kernel representing a non-differentiable random walk starting at 0.
    
    """
    assert np.all(x >= 0)
    assert np.all(y >= 0)
    return np.minimum(x, y)

@kernel(forcekron=True)
def Gibbs(x, y, scalefun=lambda x: 1):
    """
    Gibbs kernel.
    
    .. math::
        k(x, y) = \\sqrt{ \\frac {2 s(x) s(y)} {s(x)^2 + s(y)^2} }
        \\exp \\left( -\\frac {(x - y)^2} {s(x)^2 + s(y)^2} \\right),
        \\quad s = \\texttt{scalefun}.
    
    Kernel which in some sense is like a Gaussian kernel where the scale
    changes at every point. The scale is computed by the parameter `scalefun`
    which must be a callable taking the x array and returning a scale for each
    point. By default `scalefun` returns 1 so it is a Gaussian kernel.
    
    Consider that the default parameter `scale` acts before `scalefun`, so
    for example if `scalefun(x) = x` then `scale` has no effect. You should
    include all rescalings in `scalefun` to avoid surprises.
    
    """
    sx = scalefun(x)
    sy = scalefun(y)
    assert np.all(sx > 0)
    assert np.all(sy > 0)
    denom = sx ** 2 + sy ** 2
    factor = np.sqrt(2 * sx * sy / denom)
    return factor * np.exp(-(x - y) ** 2 / denom)

@stationarykernel(derivable=True, forcekron=True)
def Periodic(delta, outerscale=1):
    """
    Periodic Gaussian kernel.
    
    .. math::
        k(x, y) = \\exp \\left(
        -2 \\left(
        \\frac {\\sin((x - y) / 2)} {\\texttt{outerscale}}
        \\right)^2
        \\right)
    
    A Gaussian kernel over a transformed periodic space. It represents a
    periodic process. The usual `scale` parameter sets the period, with the
    default `scale` = 1 giving a period of 2π, while the `outerscale` parameter
    sets the length scale of the correlations.
    
    """
    assert np.isscalar(outerscale)
    assert outerscale > 0
    return np.exp(-2 * (np.sin(delta / 2) / outerscale) ** 2)

@kernel(forcekron=True, derivable=False)
def Categorical(x, y, cov=None):
    """
    Categorical kernel.
    
    .. math::
        k(x, y) = \\texttt{cov}[x, y]
    
    A kernel over integers from 0 to N-1. The parameter `cov` is the covariance
    matrix of the values.
    """
    
    # TODO support an array-like for cov, do not force it to be a numpy
    # array. In particular I'd like to support sparse matrices, but indexing
    # with two arrays is not supported for pydata/sparse. I can circumvent it
    # by flattening cov and converting x and y to flat indices manually. (Make
    # specific tests when I implement this.)
    
    assert np.issubdtype(x.dtype, numpy.integer)
    cov = np.array(cov, copy=False)
    assert len(cov.shape) == 2
    assert cov.shape[0] == cov.shape[1]
    assert np.allclose(cov, cov.T)
    return cov[x, y]

@kernel
def Rescaling(x, y, stdfun=None):
    """
    Outer product kernel.
    
    .. math::
        k(x, y) = \\texttt{stdfun}(x) \\texttt{stdfun}(y)
    
    A totally correlated kernel with arbitrary variance. Parameter `stdfun`
    must be a function that takes `x` or `y` and computes the standard
    deviation at the point. It can yield negative values; points with the same
    sign of `fun` will be totally correlated, points with different sign will
    be totally anticorrelated. Use this kernel to modulate the variance of
    other kernels. By default `stdfun` returns a constant, so it is equivalent
    to :class:`Constant`.
    
    """
    if stdfun is None:
        stdfun = lambda x: np.ones(x.shape)
        # do not use np.ones_like because it does not recognize StructuredArray
    return stdfun(x) * stdfun(y)

@stationarykernel(derivable=True, forcekron=True)
def Cos(delta):
    """
    Cosine kernel.
    
    .. math::
        k(x, y) = \\cos(x - y)
        = \\cos x \\cos y + \\sin x \\sin y
    
    Samples from this kernel are harmonic functions. It can be multiplied with
    another kernel to introduce anticorrelations.
    
    """
    return np.cos(delta)

@kernel(forcekron=True, derivable=False)
def FracBrownian(x, y, H=1/2):
    """
    Fractional brownian motion kernel.
    
    .. math::
        k(x, y) = \\frac 12 (x^{2H} + y^{2H} - |x-y|^{2H}),
        \\quad H \\in (0, 1), \\quad x, y \\ge 0
    
    For `H` = 1/2 (default) it is the Wiener kernel. For `H` in (0, 1/2) the
    increments are anticorrelated (strong oscillation), for `H` in (1/2, 1)
    the increments are correlated (tends to keep a slope).
    
    """
    
    # TODO I think the correlation between successive same step increments
    # is 2^(2H-1) - 1 in (-1/2, 1). Maybe add this to the docstring.
    
    assert np.isscalar(H)
    assert 0 < H < 1
    assert np.all(x >= 0)
    assert np.all(y >= 0)
    H2 = 2 * H
    return 1/2 * (x ** H2 + y ** H2 - np.abs(x - y) ** H2)

def _ppkernel_derivable(**kw):
    return kw.get('q', 0)

@isotropickernel(input='soft', derivable=_ppkernel_derivable)
def PPKernel(r, q=0, D=1):
    """
    Piecewise polynomial kernel.
    
    .. math::
        k(r) = \\text{polynomial}_{q,D}(r)
        \\begin{cases}
            1 - r & r \\in [0, 1)     \\\\
            0     & \\text{otherwise}
        \\end{cases}
    
    An isotropic kernel with finite support. The covariance is nonzero only
    when the distance between the points is less than 1. Parameter `q` in (0,
    1, 2, 3) sets the differentiability, while parameter `D` sets the maximum
    dimensionality the kernel can be used with. Default is `q` = 0 (non
    derivable), `D` = 1 (can be used only in 1D).
    
    """
    
    # TODO get the general formula for any q.
    
    # TODO add error checking on the dimensionality in IsotropicKernel, init
    # parameter `maxdim`.
    
    # TODO compute the kernel only on the nonzero points.
    
    # TODO find the nonzero points in O(nlogn) instead of O(n^2) by sorting
    # the inputs.
    
    assert isinstance(q, (int, np.integer))
    assert 0 <= q <= 3
    assert isinstance(D, (int, np.integer))
    assert D >= 1
    j = int(np.floor(D / 2)) + q + 1
    x = 1 - r
    if q == 0:
        poly = 1
    elif q == 1:
        poly = 1 + r * (j + 1)
    elif q == 2:
        poly = 1 + r * (j + 2 + r * ((1/3 * j +  4/3) * j + 1))
    elif q == 3:
        poly = 1 + r * (j + 3 + r * ((2/5 * j + 12/5) * j + 3 + r * (((1/15 * j + 3/5) * j + 23/15) * j + 1)))
    else:
        raise NotImplementedError
    return np.where(x > 0, x ** (j + q) * poly, 0)

@kernel(derivable=1, forcekron=True)
def WienerIntegral(x, y):
    """
    Kernel for a process whose derivative is a Wiener process.
    
    .. math::
        k(x, y) = \\frac 12 \\begin{cases}
            x^2 (y - x/3) & x < y, \\\\
            y^2 (x - y/3) & y \\le x
        \\end{cases}
    
    """
    
    # TODO can I generate this algorithmically for arbitrary integration order?
    # If I don't find a closed formula I can use sympy.
    
    # TODO write formula in terms of min(x, y) and max(x, y).
    
    assert np.all(x >= 0)
    assert np.all(y >= 0)
    return 1/2 * np.where(x < y, x**2 * (y - x/3), y**2 * (x - y/3))

@kernel(forcekron=True, derivable=True)
def Taylor(x, y):
    """
    Exponential-like power series kernel.
    
    .. math::
        k(x, y) = \\sum_{k=0}^\\infty \\frac {x^k}{k!} \\frac {y^k}{k!}
        = I_0(2 \\sqrt{xy})
    
    It is equivalent to fitting with a Taylor series expansion in zero with
    independent priors on the coefficients k with mean zero and standard
    deviation 1/k!.
    """
    # TODO what is the "natural" extension of this to multidim? Is forcekron
    # appropriate?
    
    mul = x * y
    val = 2 * np.sqrt(np.abs(mul))
    return np.where(mul >= 0, special.i0(val), special.j0(val))

@autograd.extend.primitive
def _bernoulli_poly(n, x):
    # takes x mod 1
    bernoulli = special.bernoulli(n)
    k = np.arange(n + 1)
    binom = special.binom(n, k)
    coeffs = binom[::-1] * bernoulli
    x = x % 1
    cond = x < 0.5
    x = np.where(cond, x, 1 - x)
    out = 0. * x # to handle the case n == 0
    for c in coeffs[:-1]:
        out += c
        out *= x
    out += coeffs[-1]
    if n % 2 == 1:
        out *= np.where(cond, 1, -1)
    return out

autograd.extend.defvjp(
    _bernoulli_poly,
    lambda ans, n, x: lambda g: g * (n * _bernoulli_poly(n - 1, x) if n else 0),
    argnums=[1]
)

def _fourier_derivable(**kw):
    return kw.get('n', 2) - 1

@stationarykernel(forcekron=True, derivable=_fourier_derivable, saveargs=True)
def _FourierBase(delta, n=2):
    """
    Fourier kernel.
    
    .. math::
        k(x, y) &= \\frac1{\\zeta(2n)} \\sum_{k=1}^\\infty
        \\frac {\\cos(2\\pi kx)}{k^n} \\frac {\\cos(2\\pi ky)}{k^n}
        + \\frac1{\\zeta(2n)} \\sum_{k=1}^\\infty
        \\frac {\\sin(2\\pi kx)}{k^n} \\frac {\\sin(2\\pi ky)}{k^n} = \\\\
        &= \\frac1{\\zeta(2n)} \\sum_{k=1}^\\infty
        \\frac {\\cos(2\\pi k(x-y))} {k^{2n}} = \\\\
        &= (-1)^{n+1}
        \\frac1{\\zeta(2n)} \\frac {(2\\pi)^{2n}} {2(2n)!}
        B_{2n}(x-y \\bmod 1),
    
    where :math:`B_s(x)` is a Bernoulli polynomial. It is equivalent to fitting
    with a Fourier series of period 1 with independent priors on the
    coefficients with mean zero and variance
    :math:`1/(\\zeta(2n)k^{2n})`. The process is :math:`n - 1` times
    derivable.
    
    Note that the :math:`k = 0` term is not included in the summation, so the
    mean of the process over one period is forced to be zero.
    
    """
    
    # TODO maxk parameter to truncate the series. I have to manually sum the
    # components? => Bad idea then. => Can I sum analitically the residual?
    
    # TODO add constant as option, otherwise I can't compute the Fourier
    # series when I add a constant. => Maybe this will be solved when I
    # improve the transformations system.
    
    assert isinstance(n, (int, np.integer)), type(n)
    assert n >= 1, n # TODO I could allow n == 0 to be a constant kernel
    s = 2 * n
    sign0 = -(-1) ** n
    factor = (2 * np.pi) ** s / (2 * special.factorial(s) * special.zeta(s))
    return sign0 * factor * _bernoulli_poly(s, delta)

class Fourier(_FourierBase):
    
    # TODO write a method of _KernelBase that makes a new kernel from the
    # current one to be used as starting point by all the transformation
    # methods. It should have an option on how much the subclass should be
    # preserved, for example this implementation of `fourier` is broken as soon
    # as a transformation is applied to the kernel. Linear transformations
    # should transform not only the kernel but also its transformations. I
    # have to think how to make this work in full generality. => Tentative
    # design: loc, scale, dim, etc. must become standalone methods and
    # appropriately transform the other transformation methods. Alternative:
    # interface for providing the internals of the transformations, separately
    # for cross and symmetric cases, and it's always the Kernel method that
    # manages everything like it is for __call__.
    
    def fourier(self, dox, doy):
        
        # TODO problem: this ignores completely loc and scale. Write a static
        # _KernelBase method that applies loc and scale to a kernel core, use
        # it in __init__ and here.
        
        if not dox and not doy:
            return self
        
        n = self.initargs.get('n', 2)
        s = 2 * n
        
        if dox and doy:
            def kernel(k, q):
                order = np.ceil(k / 2)
                denom = order ** s * special.zeta(s)
                return np.where((k == q) & (k > 0), 1 / denom, 0)
        
        else:
            def kernel(k, y):
                order = np.ceil(k / 2)
                denom = order ** s * special.zeta(s)
                odd = k % 2
                arg = 2 * np.pi * order * y
                return np.where(k > 0, np.where(odd, np.sin(arg), np.cos(arg)) / denom, 0)
        
            if doy:
                kernel = lambda x, q, kernel=kernel: kernel(q, x)
        
        cls = _Kernel.Kernel if dox == doy and isinstance(self, _Kernel.Kernel) else _Kernel._CrossKernel
        obj = cls(kernel, forcebroadcast=self._forcebroadcast)
        obj.initargs = self.initargs
        obj._maxderivable = self._maxderivable
        return obj

@kernel(forcekron=True, derivable=False)
def OrnsteinUhlenbeck(x, y):
    """
    Ornstein-Uhlenbeck process kernel.
    
    .. math::
        k(x, y) = \\exp(-|x - y|) - \\exp(-(x + y)),
        \\quad x, y \\ge 0
    
    It is a random walk plus a negative feedback term that keeps the
    asymptotical variance constant. It is asymptotically stationary; often the
    name "Ornstein-Uhlenbeck" is given to the stationary part only, which here
    is provided as :class:`Expon`.
    
    """
    assert np.all(x >= 0)
    assert np.all(y >= 0)
    return np.exp(-np.abs(x - y)) - np.exp(-(x + y))
    
def _Celerite_derivable(**kw):
    gamma = kw.get('gamma', 1)
    B = kw.get('B', 0)
    if np.isscalar(gamma) and np.isscalar(B) and B == gamma:
        return 1
    else:
        return False

@stationarykernel(forcekron=True, derivable=_Celerite_derivable)
def Celerite(delta, gamma=1, B=0):
    """
    Celerite kernel.
    
    .. math::
        k(x, y) = \\exp(-\\gamma|x - y|)
        \\big( \\cos(x - y) + B \\sin(|x - y|) \\big)
    
    This is the covariance function of an AR(2) process with complex roots. The
    parameters must satisfy the condition :math:`|B| \\le \\gamma`. For
    :math:`B = \\gamma` it is equivalent to the :class:`Harmonic` kernel with
    :math:`\\eta Q = 1/B, Q > 1`, and it is derivable.
    
    Reference: Daniel Foreman-Mackey, Eric Agol, Sivaram Ambikasaran, and Ruth
    Angus: *Fast and Scalable Gaussian Process Modeling With Applications To
    Astronomical Time Series*.
    """
    
    assert np.isscalar(gamma) and 0 <= gamma < np.inf
    assert np.isscalar(B) and np.isfinite(B)
    assert np.abs(B) <= gamma
    tau = _Kernel._abs(delta)
    # TODO option input='abs' in StationaryKernel
    return np.exp(-gamma * tau) * (np.cos(tau) + B * np.sin(tau))

@kernel(forcekron=True, derivable=False)
def BrownianBridge(x, y):
    """
    Brownian bridge kernel.
    
    .. math::
        k(x, y) = \\min(x, y) - xy,
        \\quad x, y \\in [0, 1]
    
    It is a Wiener process conditioned on being zero at x = 1.
    """
    
    # TODO can this have a Hurst index? I think the kernel would be
    # (t^2H(1-s) + s^2H(1-t) + s(1-t)^2H + t(1-s)^2H - (t+s) - |t-s|^2H + 2ts)/2
    # but I have to check if it is correct. (In new kernel FracBrownianBridge.)
    
    assert np.all(0 <= x) and np.all(x <= 1)
    assert np.all(0 <= y) and np.all(y <= 1)
    return np.minimum(x, y) - x * y

@stationarykernel(forcekron=True, derivable=1)
def Harmonic(delta, Q=1):
    """
    Damped stochastically driven harmonic oscillator kernel.
    
    .. math::
        k(x, y) =
        \\exp\\left( -\\frac {\\tau} {Q} \\right)
        \\begin{cases}
            \\cosh(\\eta\\tau) + \\sinh(\\eta\\tau) / (\\eta Q)
            & 0 < Q < 1 \\\\
            1 + \\tau & Q = 1 \\\\
            \\cos(\\eta\\tau) + \\sin(\\eta\\tau) / (\\eta Q)
            & Q > 1,
        \\end{cases}
    
    where :math:`\\tau = |x - y|` and :math:`\\eta = \\sqrt{|1 - 1/Q^2|}`.
    
    The process is the solution to the stochastic differential equation
    
    .. math::   f''(x) + 2/Q f'(x) + f(x) = w(x),
    
    where `w` is white noise.
    
    The parameter `Q` is the quality factor, i.e., the ratio between the energy
    stored in the oscillator and the energy lost in each cycle due to damping.
    The angular frequency is 1, i.e., the period is 2π. The process is derivable
    one time.
    
    In 1D, for `Q` = 1 (default) and `scale` = sqrt(1/3), it is the Matérn 3/2
    kernel.
    
    Reference: Daniel Foreman-Mackey, Eric Agol, Sivaram Ambikasaran, and Ruth
    Angus: *Fast and Scalable Gaussian Process Modeling With Applications To
    Astronomical Time Series*.
    """
    
    # TODO improve and test the numerical accuracy for derivatives near x=0
    # and Q=1. I don't know if the derivatives have problems away from Q=1.
    
    # TODO probably second derivatives w.r.t. Q at Q=1 are wrong.
    
    assert np.isscalar(Q) and 0 < Q < np.inf
    
    tau = _Kernel._abs(delta)
    
    if Q < 1/2:
        etaQ = np.sqrt((1 - Q) * (1 + Q))
        tauQ = tau / Q
        pexp = np.exp(_sqrt1pm1(-np.square(Q)) * tauQ)
        mexp = np.exp(-(1 + etaQ) * tauQ)
        return (pexp + mexp + (pexp - mexp) / etaQ) / 2
    
    elif 1/2 <= Q < 1:
        etaQ = np.sqrt(1 - np.square(Q))
        tauQ = tau / Q
        etatau = etaQ * tauQ
        return np.exp(-tauQ) * (np.cosh(etatau) + np.sinh(etatau) / etaQ)
        
    elif Q == 1:
        return _harmonic(tau, Q)
    
    else: # Q > 1
        etaQ = np.sqrt(np.square(Q) - 1)
        tauQ = tau / Q
        etatau = etaQ * tauQ
        return np.exp(-tauQ) * (np.cos(etatau) + np.sin(etatau) / etaQ)

def _sqrt1pm1(x):
    """sqrt(1 + x) - 1, numerically stable for small x"""
    return np.expm1(1/2 * np.log1p(x))

def _harmonic(x, Q):
    return _matern32(x / Q) + np.exp(-x/Q) * (1 - Q) * np.square(x) * (1 + x/3)

# def _harmonic(x, Q):
#     return np.exp(-x/Q) * (1 + x + (1 - Q) * x * (1 + x * (1 + x/3)))

# @autograd.extend.primitive
# def _harmonic(x, Q):
#     return (1 + x) * np.exp(-x)
#
# autograd.extend.defvjp(
#     _harmonic,
#     lambda ans, x, Q: lambda g: g * -np.exp(-x/Q) * x * (1 + (Q-1) * (1+x)),
#     lambda ans, x, Q: lambda g: g * -np.exp(-x) * x ** 3 / 3
# ) # d/dQ: -np.exp(-x/Q) * (3/Q**2 - 1) * x**3 / (6 * Q**2)
#
# autograd.extend.defjvp(
#     _harmonic,
#     lambda g, ans, x, Q: (g.T * (-np.exp(-x) * x).T).T,
#     lambda g, ans, x, Q: (g.T * (-np.exp(-x) * x ** 3 / 3).T).T
# )

@stationarykernel(forcekron=True, derivable=False)
def Expon(delta):
    """
    Exponential kernel.
    
    .. math::
        k(x, y) = \\exp(-|x - y|)
    
    In 1D it is equivalent to the Matérn 1/2 kernel, however in more dimensions
    it acts separately while the Matérn kernel is isotropic.
    """
    return np.exp(-_Kernel._abs(delta))

_bow_regexp = re.compile(r'\s|[!«»"“”‘’/()\'?¡¿„‚<>,;.:-–—]')

@kernel(forcekron=True, derivable=False)
@np.vectorize
def BagOfWords(x, y):
    """
    Bag of words kernel.
    
    .. math::
        k(x, y) &= \\sum_{w \\in \\text{words}} c_w(x) c_w(y), \\\\
        c_w(x) &= \\text{number of times word $w$ appears in $x$}
    
    The words are defined as non-empty substrings delimited by spaces or one of
    the following punctuation characters: ! « » " “ ” ‘ ’ / ( ) ' ? ¡ ¿ „ ‚ < >
    , ; . : - – —.
    """
    
    # TODO precompute the bags for x and y, then call a vectorized private
    # function.
    
    # TODO iterate on the shorter bag and use get on the other instead of
    # computing set intersection? Or: convert words to integers and then do
    # set intersection with sorted arrays?
    
    xbag = collections.Counter(_bow_regexp.split(x))
    ybag = collections.Counter(_bow_regexp.split(y))
    xbag[''] = 0
    ybag[''] = 0
    common = set(xbag) & set(ybag)
    return sum(xbag[k] * ybag[k] for k in common)



# lsqfitgp/_GP.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import sys
import builtins
import abc
import warnings

import gvar
from autograd import numpy as np
from autograd.scipy import linalg
from autograd.builtins import isinstance

from . import _Kernel
from . import _linalg
from . import _array
from . import _Deriv

__all__ = [
    'GP'
]

def _concatenate_noop(alist, **kw):
    """
    Like np.concatenate, but does not make a copy when concatenating only one
    array.
    """
    if len(alist) == 1:
        return np.array(alist[0], copy=False)
    else:
        return np.concatenate(alist, **kw)

def _triu_indices_and_back(n):
    """
    Return indices to get the upper triangular part of a matrix, and indices to
    convert a flat array of upper triangular elements to a symmetric matrix.
    """
    indices = np.triu_indices(n)
    q = np.empty((n, n), int)
    a = np.arange(len(indices[0]))
    q[indices] = a
    q[tuple(reversed(indices))] = a
    return indices, q

def _block_matrix(blocks):
    """
    Like np.block, but is autograd-friendly and avoids a copy when there is
    only one block.
    """
    return _concatenate_noop([_concatenate_noop(row, axis=1) for row in blocks], axis=0)
    # TODO make a bug report to autograd because np.block does not work

def _isarraylike_nostructured(x):
    return np.isscalar(x) or isinstance(x, (list, tuple, np.ndarray))

def _isarraylike(x):
    return _isarraylike_nostructured(x) or isinstance(x, _array.StructuredArray)

def _isdictlike(x):
    return isinstance(x, (dict, gvar.BufferDict))

def _compatible_dtypes(d1, d2): # pragma: no cover
    """
    Function to check x arrays datatypes passed to GP.addx. If the dtype is
    structured, it checks the structure of the fields is the same, but allows
    casting of concrete dtypes (like, in one array a field can be int, in
    another float, as long as the field name and position is the same).
    Currently not used.
    """
    if d1.names != d2.names or d1.shape != d2.shape:
        return False
    if d1.names is not None:
        for name in d1.names:
            if not _compatible_dtypes(d1.fields[name][0], d2.fields[name][0]):
                return False
    else:
        try:
            np.result_type(d1, d2)
        except TypeError:
            return False
    return True

def _array_copy_if_not_readonly(x): # pragma: no cover
    """currently not used"""
    y = _array.asarray(x)
    if x is y and builtins.isinstance(x, np.ndarray) and x.flags['WRITEABLE']:
        return np.copy(x)
    else:
        return y

class _Element(metaclass=abc.ABCMeta):
    """
    Abstract class for an object holding information associated to a key in a
    GP object.
    """
    
    @property
    @abc.abstractmethod
    def shape(self): # pragma: no cover
        """Output shape"""
        pass
    
    @property
    def size(self):
        return np.prod(self.shape, dtype=int)

class _Points(_Element):
    """Points where the process is evaluated"""
    
    def __init__(self, x, deriv, proc):
        assert _isarraylike(x)
        assert isinstance(deriv, _Deriv.Deriv)
        self.x = x
        self.deriv = deriv
        self.proc = proc
    
    @property
    def shape(self):
        return self.x.shape

class _Transf(_Element):
    """Trasformation over other _Element objects"""
    
    def __init__(self, tensors, shape, axes):
        assert isinstance(tensors, dict)
        assert isinstance(shape, tuple)
        self.tensors = tensors # dict key -> array
        self._shape = shape
        self._axes = axes
    
    @property
    def shape(self):
        return self._shape
    
    def tensormul(self, tensor, x):
        """
        Do the multiplication tensor @ x.
        """
        if tensor.shape:
            return np.tensordot(tensor, x, axes=self._axes)
        else:
            return tensor * x

class _Cov(_Element):
    """User-provided covariance matrix block(s)"""
    
    def __init__(self, blocks, shape):
        """blocks = dict (key, key) -> matrices"""
        self.blocks = blocks
        self._shape = shape
    
    @property
    def shape(self):
        return self._shape

class _Proc(metaclass=abc.ABCMeta):
    """
    Abstract base class for an object holding information about a process
    in a GP object.
    """
    
    @abc.abstractmethod
    def __init__(self): # pragma: no cover
        pass
    
class _ProcKernel(_Proc):
    """An independent process defined with a kernel"""
    
    def __init__(self, kernel, deriv=0):
        assert isinstance(kernel, _Kernel.Kernel)
        self.kernel = kernel
        self.deriv = deriv
    
class _ProcTransf(_Proc):
    """A process defined as a linear transformation of other processes"""
    
    def __init__(self, ops, deriv):
        """ops = dict proc key -> callable"""
        self.ops = ops
        self.deriv = deriv

class _ProcKernelOp(_Proc):
    """A process defined by an operation on the kernel of another process"""
    
    def __init__(self, proc, method, arg):
        """proc = proc key, method = Kernel method, arg = argument to method"""
        self.proc = proc
        self.method = method
        self.arg = arg
        
class DefaultProcess:
    """Object used as key for the default process in GP objects"""
    pass

class _ZeroKernel:
    """Object representing a null kernel"""
    pass

class GP:
    """
    
    Object that represents a Gaussian process over arbitrary input.
    
    Methods
    -------
    addx
        Add points where the Gaussian process is evaluated.
    addtransf
        Define a linear transformation of the evaluated process.
    addcov
        Introduce a set of user-provided prior covariance matrix blocks.
    addproc
        Define a new independent component of the process.
    addproctransf
        Define a linear transformation of the process.
    addkernelop
        Define a transformation of the process through a kernel method.
    addprocderiv
        Define a derivative of the process.
    addprocxtransf
        Define a process with transformed inputs.
    addprocrescale
        Define a rescaled process.
    prior
        Compute the prior for the process.
    pred
        Compute the posterior for the process.
    predfromfit
        Like `pred` with ``fromdata=False``.
    predfromdata
        Like `pred` with ``fromdata=True``.
    marginal_likelihood
        Compute the "marginal likelihood", also known as "bayes factor".
    
    Parameters
    ----------
    covfun : Kernel or None
        An instance of `Kernel` representing the covariance kernel of the
        default process of the GP object. It can be left unspecified.
    solver : str
        The algorithm used to decompose the prior covariance matrix. See list
        below for the available solvers. Default is `eigcut+` which is slow but
        robust.
        
        'eigcut+'
            Promote small (or negative) eigenvalues to a minimum value.
        'eigcut-'
            Remove small (or negative) eigenvalues.
        'svdcut+'
            Promote small eigenvalues to a minimum value, keeping their sign.
        'svdcut-'
            Remove small eigenvalues.
        'lowrank'
            Reduce the rank of the matrix. The complexity is O(n^2 r) where
            `n` is the matrix size and `r` the required rank, while other
            algorithms are O(n^3). Slow for small sizes.
        'chol'
            Cholesky decomposition after regularizing the matrix with a
            Gershgorin estimate of the maximum eigenvalue. The fastest of the
            O(n^3) algorithms.
    
    checkpos : bool
        If True (default), raise a `LinAlgError` if the prior covariance
        matrix turns out non positive within numerical error. The check
        will be done only if you use in some way the `gvar` prior. With
        the Cholesky solvers the check will be done in all cases.
    checksym : bool
        If True (default), check that the prior covariance matrix is
        symmetric. If False, only half of the matrix is computed.
    checkfinite : bool
        If True (default), check that the prior covariance matrix does not
        contain infs or nans.
    **kw
        Additional keyword arguments are passed to the solver.
        
        eps : positive float
            For solvers 'eigcut+', 'eigcut-', 'svdcut+', 'svdcut-', 'chol'.
            Specifies the threshold for considering small the eigenvalues,
            relative to the maximum eigenvalue. The default is matrix size *
            float epsilon.
        rank : positive integer
            For the 'lowrank' solver, the target rank. It should be much
            smaller than the matrix size for the method to be convenient.

    """
    
    # TODO maybe checkpos=True is a too strong default. It can quickly grow
    # beyond doable even if the actual computations remain light. Also, it
    # does not really make sense that it is used only in case of gvar output.
    # Alternative 1: set checkpos=False by default and if True build all the
    # covariance blocks the first time I call _covblock and check everything.
    # Alternative 2: only check the matrices I need to invert. Alternative 3:
    # the default is to check only if the matrices are small enough that it is
    # not a problem. Alternative 4: check diagonal blocks.
    # Current behaviour: only the covariance of things added with addx is
    # checked, and only if gvars are used at some point
    # Alternative 5: do the check in _solver only on things to be solved
    # => Idea: do the full check even when not using gvars, only on
    # non-transformed variables, but issue a warning if the system is large
    
    # TODO maybe the default should be solver='eigcut-' because it adds less
    # noise. It is moot anyway if the plot is done sampling with gvar.raniter
    # or lgp.raninter because they do add their own noise. But this will be
    # solved when I add GP sampling methods.
    
    def __init__(self, covfun=None, solver='eigcut+', checkpos=True, checksym=True, checkfinite=True, **kw):
        self._elements = dict() # key -> _Element
        self._covblocks = dict() # (key, key) -> matrix
        self._decompcache = dict() # tuple of keys -> Decomposition
        self._procs = dict() # proc key -> _Proc
        self._kernels = dict() # (proc key, proc key) -> _KernelBase
        if covfun is not None:
            if not isinstance(covfun, _Kernel.Kernel):
                raise TypeError('covariance function must be of class Kernel')
            self._procs[DefaultProcess] = _ProcKernel(covfun)
        self._canaddx = True
        decomp = {
            'eigcut+': _linalg.EigCutFullRank,
            'eigcut-': _linalg.EigCutLowRank,
            'svdcut+': _linalg.SVDCutFullRank,
            'svdcut-': _linalg.SVDCutLowRank,
            'lowrank': _linalg.ReduceRank,
            'chol'   : _linalg.CholGersh,
        }[solver]
        self._decompclass = lambda K, **kwargs: decomp(K, **kwargs, **kw)
        self._checkpositive = bool(checkpos)
        self._checksym = bool(checksym)
        self._checkfinite = bool(checkfinite)
    
    def addproc(self, kernel=None, key=DefaultProcess, deriv=0):
        """
        
        Add an independent process.
        
        Parameters
        ----------
        kernel : Kernel
            A kernel for the process. If None, use the default kernel. The
            difference between the default process and a process defined with
            the default kernel is that, although they have the same kernel,
            they are independent.
        key : hashable
            The name that identifies the process in the GP object. If not
            specified, sets the kernel of the default process.
        deriv : Deriv-like
            Derivatives to take on the process defined by the kernel.
                
        """
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        if kernel is None:
            kernel = self._procs[DefaultProcess].kernel
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = _ProcKernel(kernel, deriv)
    
    def addproctransf(self, ops, key, deriv=0):
        """
        
        Define a new process as a linear combination of other processes.
        
        Let f_i(x), i = 1, 2, ... be already defined processes, and g_i(x) be
        deterministic functions. The new process is defined as
        
            h(x) = g_1(x) f_1(x) + g_2(x) f_2(x) + ...
        
        Parameters
        ----------
        ops : dict
            A dictionary mapping process keys to scalars or scalar
            functions. The functions must take an argument of the same kind
            of the domain of the process.
        key : hashable
            The name that identifies the new process in the GP object.
        deriv : Deriv-like
            The linear combination is derived as specified by this
            parameter.
        
        """
        
        for k, func in ops.items():
            if k not in self._procs:
                raise KeyError(f'process key {k!r} not in GP object')
            if not np.isscalar(func) and not callable(func):
                raise TypeError(f'object of type {type(func)!r} for key {k!r} is neither scalar nor callable')
        
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        
        deriv = _Deriv.Deriv(deriv)
        
        self._procs[key] = _ProcTransf(ops, deriv)
    
    def addkernelop(self, method, arg, key, proc=DefaultProcess):
        """
        
        Define a new process as the transformation of an existing one.
        
        Parameters
        ----------
        method : str
            A method of `Kernel` taking two arguments which returns a
            transformed kernel.
        arg : object
            A valid argument to the method.
        key : hashable
            Key for the new process.
        proc : hashable
            Key of the process to be transformed. If not specified, use the
            default process.
        
        """
        
        if not hasattr(_Kernel.Kernel, method):
            raise ValueError(f'Kernel has not attribute {method!r}')
        if key in self._procs:
            raise KeyError(f'process key {key!r} already used in GP')
        if proc not in self._procs:
            raise KeyError(f'process {proc!r} not found')
                
        self._procs[key] = _ProcKernelOp(proc, method, arg)
    
    def addprocderiv(self, deriv, key, proc=DefaultProcess):
        """
        
        Define a new process as the derivative of an existing one.
        
        .. math::
            g(x) = \\frac{\\partial^n}{\\partial x^n} f(x)
        
        Parameters
        ----------
        deriv : Deriv-like
            Derivation order.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be derived. If not specified, use the
            default process.
        
        """
        deriv = _Deriv.Deriv(deriv)
        self.addkernelop('diff', deriv, key, proc)
    
    def addprocxtransf(self, transf, key, proc=DefaultProcess):
        """
        
        Define a new process by transforming the inputs of another one.
        
        .. math::
            g(x) = f(T(x))
        
        Parameters
        ----------
        transf : callable
            A function mapping the new kind input to the input expected by the
            transformed process.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be transformed. If not specified, use the
            default process.
        
        """
        assert callable(transf)
        self.addkernelop('xtransf', transf, key, proc)
    
    def addprocrescale(self, scalefun, key, proc=DefaultProcess):
        """
        
        Define a new process as a rescaling of an existing one.
        
        .. math::
            g(x) = s(x)f(x)
        
        Parameters
        ----------
        scalefun : callable
            A function from the domain of the process to a scalar.
        key : hashable
            The key of the new process.
        proc : hashable
            The key of the process to be transformed. If not specified, use the
            default process.
        
        """
        assert callable(scalefun)
        self.addkernelop('rescale', scalefun, key, proc)
    
    def addx(self, x, key=None, deriv=0, proc=DefaultProcess):
        """
        
        Add points where the Gaussian process is evaluated.
        
        The GP object keeps the various x arrays in a dictionary. If `x` is an
        array, you have to specify its dictionary key with the `key` parameter.
        Otherwise, you can directly pass a dictionary for `x`.
        
        To specify that on the given `x` a derivative of the process instead of
        the process itself should be evaluated, use the parameter `deriv`.
        
        `addx` never copies the input arrays if they are numpy arrays, so if
        you change their contents before doing something with the GP, the
        change will be reflected on the result. However, after the GP has
        computed internally its covariance matrix, the x are ignored.
        
        If you use in some way the `gvar` prior, e.g., by calling `prior` or
        `pred` using gvars, you can't call `addx` any more, due to a
        limitation in gvar.
        
        Parameters
        ----------
        x : array or dictionary of arrays
            The points to be added.
        key : hashable
            If `x` is an array, the dictionary key under which `x` is added.
            Can not be specified if `x` is a dictionary.
        deriv : Deriv-like
            Derivative specification. A :class:`Deriv` object or something that
            can be converted to :class:`Deriv`.
        proc : hashable
            The process to be evaluated on the points. If not specified, use
            the default process.
        
        """
        
        # TODO after I implement block solving, add per-key solver option. It
        # may not be possible to respect the setting if it is not the sole
        # key used as data.
        
        # TODO add `copy` parameter, default False, to copy the input arrays.
        
        if not self._canaddx:
            raise RuntimeError('can not add x any more to this process')
            # TODO remove if I implement the lazy gvar prior.
        
        deriv = _Deriv.Deriv(deriv)

        if proc not in self._procs:
            raise KeyError(f'process named {proc!r} not found')
                
        if _isarraylike(x):
            if key is None:
                raise ValueError('x is array but key is None')
            x = {key: x}
        elif _isdictlike(x):
            if key is not None:
                raise ValueError('can not specify key if x is a dictionary')
            if None in x:
                raise ValueError('None key in x not allowed')
        else:
            raise TypeError('x must be array or dict')
        
        for key in x:
            if key in self._elements:
                raise KeyError('key {!r} already in GP'.format(key))
            
            gx = x[key]
            
            # Convert to numpy array or StructuredArray.
            if not _isarraylike(gx):
                raise TypeError('x[{!r}] is not array-like'.format(key))
            gx = _array.asarray(gx)

            # Check it is not empty.
            if not gx.size:
                raise ValueError('x[{!r}] is empty'.format(key))
                # TODO it would be elegant to support empty arrays, but how
                # many things would break down from here?

            # Check dtype is compatible with previous arrays.
            # TODO since we never concatenate arrays we could allow a less
            # strict compatibility. In principle we could allow really anything
            # as long as the kernel eats it, but this probably would let bugs
            # through without being really ever useful. What would make sense
            # is checking the dtype structure matches recursively and check
            # concrete dtypes of fields can be casted.
            # TODO result_type is too lax. Examples: str, float -> str,
            # object, float -> object. I should use something like the
            # ordering function in updowncast.py.
            if hasattr(self, '_dtype'):
                try:
                    self._dtype = np.result_type(self._dtype, gx.dtype)
                except TypeError:
                    msg = 'x[{!r}].dtype = {!r} not compatible with {!r}'
                    msg = msg.format(key, gx.dtype, self._dtype)
                    raise TypeError(msg)
            else:
                self._dtype = gx.dtype

            # Check that the derivative specifications are compatible with the
            # array data type.
            if gx.dtype.names is None:
                if not deriv.implicit:
                    raise ValueError('x has not fields but derivative has')
            else:
                for dim in deriv:
                    if dim not in gx.dtype.names:
                        raise ValueError(f'deriv field {dim!r} not in x')
            
            self._elements[key] = _Points(gx, deriv, proc)
        
    def addtransf(self, tensors, key, axes=1):
        """
        
        Apply a linear transformation to already specified process points. The
        result of the transformation is represented by a new key.
        
        Parameters
        ----------
        tensors : dict
            Dictionary mapping keys of the GP to arrays/scalars. Each array is
            matrix-multiplied with the process array represented by its key,
            while scalars are just multiplied. Finally, the keys are summed
            over.
        key : hashable
            A new key under which the transformation is placed.
        axes : int
            Number of axes to be summed over for matrix multiplication,
            referring to trailing axes for tensors in `tensors`, and to
            heading axes for process points. Default 1.
        
        Notes
        -----
        The multiplication between the tensors and the process is done with
        np.tensordot with, by default, 1-axis contraction. For >2d arrays this
        is different from numpy's matrix multiplication, which would act on the
        second-to-last dimension of the second array.
        
        """        
        # TODO if a tensor is a callable, it is called on the points to get
        # the tensor. It can be only applied on _Points. The callable is called
        # immediately in addtransf to catch errors.
        
        # TODO with callable transf, it would be nice to take derivatives
        # on the combined transf @ kernel @ transf.T. Maybe then it is better
        # to add transformations as a kernel method.
        
        # Note: it may seem nice that when an array has less axes than `axes`,
        # the summation would be restricted only on the existing axes. However
        # this brings about the ambiguous case where only one of the factors has
        # not enough axes. How many axes do you sum over on the other?
        
        # Check axes.
        assert isinstance(axes, int), axes
        assert axes >= 1, axes
        
        # Check key.
        if key is None:
            raise ValueError('key can not be None')
        if key in self._elements:
            raise KeyError(f'key {key!r} already in GP')
        
        # Check keys.
        for k in tensors:
            if k not in self._elements:
                raise KeyError(k)
        
        # Check tensors and convert them to numpy arrays.
        tens = {}
        for k, t in tensors.items():
            t = np.array(t, copy=False)
            if not np.issubdtype(t.dtype, np.number):
                raise TypeError(f'tensors[{k!r}] has non-numeric dtype {t.dtype!r}')
            if self._checkfinite and not np.all(np.isfinite(t)):
                raise ValueError(f'tensors[{k!r}] contains infs/nans')
            rshape = self._elements[k].shape
            if t.shape and t.shape[-axes:] != rshape[:axes]:
                raise ValueError(f'tensors[{k!r}].shape = {t.shape!r} can not be multiplied with shape {rshape!r} with {axes}-axes contraction')
            tens[k] = t
        
        # Compute shape.
        arrays = tens.values()
        elements = (self._elements[k] for k in tens)
        shapes = [
            t.shape[:-axes] + e.shape[axes:] if t.shape else e.shape
            for t, e in zip(arrays, elements)
        ]
        try:
            shape = _array.broadcast_shapes(shapes)
        except ValueError:
            msg = 'can not broadcast tensors with shapes ['
            msg += ', '.join(repr(t.shape) for t in arrays)
            msg += '] contracted with arrays with shapes ['
            msg += ', '.join(repr(e.shape) for e in elements) + ']'
            raise ValueError(msg)
        
        self._elements[key] = _Transf(tens, shape, axes)
    
    def addcov(self, covblocks, key=None):
        """
        
        Add user-defined prior covariance matrix blocks.
        
        Covariance matrices defined with `addcov` represent arbitrary finite
        zero-mean Gaussian variables, assumed independent from all other
        variables in the GP object.
                
        Parameters
        ----------
        covblocks : array or dictionary of arrays
            If an array: a covariance matrix (or tensor) to be added under key
            `key`. If a dictionary: a mapping from pairs of keys to the
            corresponding covariance matrix blocks. A missing off-diagonal
            block in the dictionary is interpreted as a matrix of zeros.
        key : hashable
            If `covblocks` is an array, the dictionary key under which
            `covblocks` is added. Can not be specified if `covblocks` is a
            dictionary.
        
        Raises
        ------
        RuntimeError :
            A key is already used in the GP.
        ValueError :
            `covblocks` and/or `key` are malformed or inconsistent.
        
        """
        
        if not self._canaddx:
            raise RuntimeError('can not add x any more to this process')
            # TODO remove if I implement the lazy gvar prior.
        
        # Check type of `covblocks` and standardize it to dictionary.
        if _isarraylike(covblocks):
            if key is None:
                raise ValueError('covblocks is array but key is None')
            covblocks = {(key, key): covblocks}
        elif _isdictlike(covblocks):
            if key is not None:
                raise ValueError('can not specify key if covblocks is a dictionary')
            if None in covblocks:
                raise ValueError('None key in covblocks not allowed')
        else:
            raise TypeError('covblocks must be array or dict')
        
        # Convert blocks to numpy arrays and determine shapes from diagonal
        # blocks.
        shapes = {}
        preblocks = {}
        for keys, block in covblocks.items():
            for key in keys:
                if key in self._elements:
                    raise KeyError(f'key {key!r} already in GP')
            xkey, ykey = keys
            block = np.asarray(block)
            if xkey == ykey:
                if block.ndim % 2 == 1:
                    raise ValueError(f'diagonal block {key!r} has odd number of axes')
                half = block.ndim // 2
                head = block.shape[:half]
                tail = block.shape[half:]
                if head != tail:
                    raise ValueError(f'shape {block.shape!r} of diagonal block {key!r} is not symmetric')
                if self._checksym:
                    if not np.allclose(block, block.T):
                        raise ValueError(f'diagonal block {key!r} is not symmetric')
                shapes[xkey] = head
            preblocks[keys] = block
        
        # Reshape blocks to square matrices and check that the shapes of out of
        # diagonal blocks match those of diagonal ones.
        blocks = {}
        for keys, block in preblocks.items():
            if self._checkfinite and not np.all(np.isfinite(block)):
                raise ValueError(f'block {keys!r} not finite')
            xkey, ykey = keys
            if xkey == ykey:
                size = np.prod(shapes[xkey], dtype=int)
                blocks[keys] = block.reshape((size, size))
            else:
                for key in keys:
                    if key not in shapes:
                        raise KeyError(f'key {key!r} from off-diagonal block {keys!r} not found in diagonal blocks')
                eshape = shapes[xkey] + shapes[ykey]
                if block.shape != eshape:
                    raise ValueError(f'shape {block.shape!r} of block {keys!r} is not {eshape!r} as expected from diagonal blocks')
                xsize = np.prod(shapes[xkey], dtype=int)
                ysize = np.prod(shapes[ykey], dtype=int)
                block = block.reshape((xsize, ysize))
                blocks[keys] = block
                revkeys = (ykey, xkey)
                blockT = preblocks.get(revkeys)
                if blockT is None:
                    blocks[revkeys] = block.T
        
        # Check symmetry of out of diagonal blocks.
        if self._checksym:
            for keys, block in blocks.items():
                xkey, ykey = keys
                if xkey != ykey:
                    blockT = blocks[ykey, xkey]
                    if not np.allclose(block.T, blockT):
                        raise ValueError(f'block {keys!r} is not the transpose of block {revkeys!r}')
        
        # Create _Cov objects.
        for key, shape in shapes.items():
            self._elements[key] = _Cov(blocks, shape)

    def _crosskernel(self, xpkey, ypkey):
        
        # Check if the kernel is in cache.
        cache = self._kernels.get((xpkey, ypkey))
        if cache is not None:
            return cache
        
        # Compute the kernel.
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if isinstance(xp, _ProcKernel) and isinstance(yp, _ProcKernel):
            kernel = self._crosskernel_kernels(xpkey, ypkey)
        elif isinstance(xp, _ProcTransf):
            kernel = self._crosskernel_transf_any(xpkey, ypkey)
        elif isinstance(yp, _ProcTransf):
            kernel = self._crosskernel_transf_any(ypkey, xpkey)
            kernel = kernel if kernel is _ZeroKernel else kernel._swap()
        elif isinstance(xp, _ProcKernelOp):
            kernel = self._crosskernel_op_any(xpkey, ypkey)
        elif isinstance(yp, _ProcKernelOp):
            kernel = self._crosskernel_op_any(ypkey, xpkey)
            kernel = kernel if kernel is _ZeroKernel else kernel._swap()
        else:
            raise TypeError(f'unrecognized process types {type(xp)!r} and {type(yp)!r}')
        
        # Save cache.
        self._kernels[xpkey, ypkey] = kernel
        self._kernels[ypkey, xpkey] = kernel if kernel is _ZeroKernel else kernel._swap()
        
        return kernel
        
    def _crosskernel_kernels(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            return xp.kernel.diff(xp.deriv, xp.deriv)
        else:
            return _ZeroKernel
    
    def _crosskernel_transf_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        kernelsum = _ZeroKernel
        
        for pkey, factor in xp.ops.items():
            kernel = self._crosskernel(pkey, ypkey)
            if kernel is _ZeroKernel:
                continue
            
            if np.isscalar(factor):
                factor = lambda x, factor=factor: factor
            kernel = kernel.rescale(factor, None)
            
            if kernelsum is _ZeroKernel:
                kernelsum = kernel
            else:
                kernelsum += kernel
        
        if kernelsum is not _ZeroKernel:
            kernelsum = kernelsum.diff(xp.deriv, 0)
        
        return kernelsum
    
    def _crosskernel_op_any(self, xpkey, ypkey):
        xp = self._procs[xpkey]
        yp = self._procs[ypkey]
        
        if xp is yp:
            basekernel = self._crosskernel(xp.proc, xp.proc)
            # In principle I could avoid handling this case separately but it
            # will probably allow simplifications in how I implement nontrivial
            # transformations in Kernel, I won't need to support taking a
            # transformation in two steps.
        else:
            basekernel = self._crosskernel(xp.proc, ypkey)
        
        if basekernel is _ZeroKernel:
            return _ZeroKernel
        elif xp is yp:
            return getattr(basekernel, xp.method)(xp.arg, xp.arg)
        else:
            return getattr(basekernel, xp.method)(xp.arg, None)
    
    def _makecovblock_points(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        
        assert isinstance(x, _Points)
        assert isinstance(y, _Points)
        
        kernel = self._crosskernel(x.proc, y.proc)
        if kernel is _ZeroKernel:
            # TODO handle zero cov block efficiently
            return np.zeros((x.size, y.size))
        
        kernel = kernel.diff(x.deriv, y.deriv)
        
        if x is y and not self._checksym:
            indices, back = _triu_indices_and_back(x.size)
            ax = np.broadcast_to(x.x.reshape(-1)[:, None], (x.x.size, y.x.size))[indices]
            ay = np.broadcast_to(y.x.reshape(-1)[None, :], (x.x.size, y.x.size))[indices]
            halfcov = kernel(ax, ay)
            cov = halfcov[back]
        else:
            ax = x.x.reshape(-1)[:, None]
            ay = y.x.reshape(-1)[None, :]
            cov = kernel(ax, ay)
        
        return cov
    
    def _makecovblock_transf_any(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        assert isinstance(x, _Transf)
        covsum = None
        for key, tensor in x.tensors.items():
            elem = self._elements[key]
            cov = self._covblock(key, ykey)
            assert cov.shape == (elem.size, y.size)
            cov = cov.reshape(elem.shape + y.shape)
            cov = x.tensormul(tensor, cov)
            if covsum is not None:
                covsum = covsum + cov
            else:
                covsum = cov
        assert covsum.shape == x.shape + y.shape
        return covsum.reshape((x.size, y.size)) # !don't leave out the ()!
            
    def _makecovblock(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        if isinstance(x, _Points) and isinstance(y, _Points):
            cov = self._makecovblock_points(xkey, ykey)
        elif isinstance(x, _Transf):
            cov = self._makecovblock_transf_any(xkey, ykey)
        elif isinstance(y, _Transf):
            cov = self._makecovblock_transf_any(ykey, xkey)
            cov = cov.T
        elif isinstance(x, _Cov) and isinstance(y, _Cov) and x.blocks is y.blocks and (xkey, ykey) in x.blocks:
            cov = x.blocks[xkey, ykey]
        else:
            # TODO handle zero cov block efficiently
            cov = np.zeros((x.size, y.size))

        if self._checkfinite and not np.all(np.isfinite(cov)):
            raise RuntimeError('covariance block {!r} is not finite'.format((xkey, ykey)))
        if self._checksym and xkey == ykey and not np.allclose(cov, cov.T):
            raise RuntimeError('covariance block {!r} is not symmetric'.format((xkey, ykey)))

        return cov

    def _covblock(self, row, col):        
        if (row, col) not in self._covblocks:
            block = self._makecovblock(row, col)
            _linalg.noautograd(block).flags['WRITEABLE'] = False
            if row != col:
                if self._checksym:
                    blockT = self._makecovblock(col, row)
                    if not np.allclose(block.T, blockT):
                        msg = 'covariance block {!r} is not symmetric'
                        raise RuntimeError(msg.format((row, col)))
                self._covblocks[col, row] = block.T
            self._covblocks[row, col] = block
        
        return self._covblocks[row, col]
        
    def _assemblecovblocks(self, rowkeys, colkeys=None):
        if colkeys is None:
            colkeys = rowkeys
        blocks = [
            [self._covblock(row, col) for col in colkeys]
            for row in rowkeys
        ]
        return _block_matrix(blocks)
    
    def _solver(self, keys, ycov=None):
        """
        Return a decomposition of the covariance matrix of the keys in `keys`
        plus the matrix ycov.
        """
        
        # TODO Block matrix solving. Example: solve a subproblem with kronecker,
        # another plain.
        
        # TODO Cache decompositions of blocks. Caching is effective
        # with data if I can reuse the decomposition of Kxx to compute the
        # decomposition of Kxx + ycov, i.e., it works in all cases if ycov is
        # scalar, and in some cases if ycov is diagonal. Is there an efficient
        # way to update a Cholesky decomposition if I add a diagonal matrix?
        
        # TODO Check if the matrix is block diagonal. It's O(n^2). It is often
        # not needed but when it is it makes a difference. A faster and less
        # thorough check can be done only on off-diagonal key-key blocks being
        # zero, which may be useful with multi-output or split components.
        
        keys = tuple(keys)
        
        # Check if decomposition is in cache.
        if ycov is None:
            cache = self._decompcache.get(keys)
            if cache is not None:
                return cache
            # TODO use frozenset(keys) instead of tuple(keys) to make cache
            # work when order changes, but I have to permute the decomposition
            # to make that work. Needs an ad-hoc class in _linalg.
        
        # Compute decomposition.
        Kxx = self._assemblecovblocks(keys)
        if ycov is not None:
            Kxx = Kxx + ycov
        decomp = self._decompclass(Kxx)
        
        # Cache decomposition.
        if ycov is None:
            self._decompcache[keys] = decomp
        
        return decomp
        
    def _checkpos(self, cov):
        # faster but less trivial options for checking positivity:
        # 1) cholesky + eps, if fails it's not positive
        # 2) ldlt, check each 2x2 block     <--- probably best? is it stable?
        # 3) QR, check diagonal of R
        eigv = linalg.eigvalsh(_linalg.noautograd(cov))
        mineigv = np.min(eigv)
        if mineigv < 0:
            bound = -len(cov) * np.finfo(float).eps * np.max(eigv)
            if mineigv < bound:
                msg = 'covariance matrix is not positive definite: '
                msg += 'mineigv = {:.4g} < {:.4g}'.format(mineigv, bound)
                raise np.linalg.LinAlgError(msg)
    
    def _priorpointscov(self, key):
        
        # TODO I think that gvar internals would let me build the prior
        # one block at a time although everything is correlated, but I don't
        # know how to do it. If I do that, remove _canaddx. => Lepage is
        # working on it.
        
        assert not hasattr(self, '_priordict')
        # caching is managed by _prior and since we have to create all gvars
        # at once this method can be called only the first time
        
        if self._checkpositive:
            keys = [
                k for k, v in self._elements.items()
                if isinstance(v, (_Points, _Cov))
            ]
            fullcov = self._assemblecovblocks(list(keys))
            self._checkpos(fullcov)
        items = [
            (k, v) for k, v in self._elements.items()
            if isinstance(v, (_Points, _Cov))
        ]
        mean = {
            key: np.zeros(x.shape)
            for key, x in items
        }
        cov = {
            (row, col): self._covblock(row, col).reshape(x.shape + y.shape)
            for row, x in items
            for col, y in items
        }
        self._priordict = gvar.gvar(mean, cov, fast=True)
        self._priordict.buf.flags['WRITEABLE'] = False
        self._canaddx = False
        
        return self._priordict[key]
    
    def _priortransf(self, key):
        x = self._elements[key]
        assert isinstance(x, _Transf)
        out = None
        for k, tensor in x.tensors.items():
            prior = self._prior(k)
            transf = x.tensormul(tensor, prior)
            if out is None:
                out = transf
            else:
                out = out + transf
        return out
    
    def _prior(self, key):
        prior = getattr(self, '_priordict', {}).get(key, None)
        if prior is None:
            x = self._elements[key]
            if isinstance(x, (_Points, _Cov)):
                prior = self._priorpointscov(key)
                # _priorpoints already caches the result
            elif isinstance(x, _Transf):
                prior = self._priortransf(key)
                self._priordict[key] = prior
            else:
                raise TypeError(type(x))
        return prior
    
    def prior(self, key=None, raw=False):
        """
        
        Return an array or a dictionary of arrays of gvars representing the
        prior for the Gaussian process. The returned object is not unique but
        the gvars stored inside are, so all the correlations are kept between
        objects returned by different calls to `prior`.
        
        Calling without arguments returns the complete prior as a dictionary.
        If you specify `key`, only the array for the requested key is returned.
        
        Parameters
        ----------
        key : None, key or list of keys
            Key(s) corresponding to one passed to `addx` or `addtransf`. None
            for all keys.
        raw : bool
            If True, instead of returning a collection of gvars return
            their covariance matrix as would be returned by `gvar.evalcov`.
            Default False.
        
        Returns
        -------
        If raw=False (default):
        
        prior : np.ndarray or gvar.BufferDict
            A collection of gvars representing the prior.
        
        If raw=True:
        
        cov : np.ndarray or dict
            The covariance matrix of the prior.
        """
        raw = bool(raw)
        
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = None
        
        if raw and outkeys is not None:
            return {
                (row, col):
                self._covblock(row, col).reshape(self._elements[row].shape + self._elements[col].shape)
                for row in outkeys
                for col in outkeys
            }
        elif raw:
            return self._covblock(key, key).reshape(2 * self._elements[key].shape)
        elif outkeys is not None:
            return gvar.BufferDict({
                key: self._prior(key) for key in outkeys
            })
        else:
            return self._prior(key)
        
    def _flatgiven(self, given, givencov):
        
        if not hasattr(given, 'keys'):
            raise TypeError('`given` must be dict')
        if givencov is not None and not hasattr(givencov, 'keys'):
            raise TypeError('`givenconv` must be None or dict')
        
        ylist = []
        keylist = []
        for key, l in given.items():
            if key not in self._elements:
                raise KeyError(key)

            l = np.array(l, copy=False)
            shape = self._elements[key].shape
            if l.shape != shape:
                msg = 'given[{!r}] has shape {!r} different from shape {!r}'
                raise ValueError(msg.format(key, l.shape, shape))
            if l.dtype != object and not np.issubdtype(l.dtype, np.number):
                msg = 'given[{!r}] has non-numerical dtype {!r}'
                raise TypeError(msg.format(key, l.dtype))
            
            ylist.append(l.reshape(-1))
            keylist.append(key)
        
        # TODO error checking on the unpacking of givencov
        
        if givencov is None:
            covblocks = None
        else:
            covblocks = [
                [
                    givencov[keylist[i], keylist[j]].reshape(ylist[i].shape + ylist[j].shape)
                    for j in range(len(keylist))
                ]
                for i in range(len(keylist))
            ]
            
        return ylist, keylist, covblocks
    
    def _slices(self, keylist):
        """
        Return list of slices for the positions of flattened arrays
        corresponding to keys in `keylist` into their concatenation.
        """
        sizes = [self._elements[key].size for key in keylist]
        stops = np.concatenate([[0], np.cumsum(sizes)])
        return [slice(stops[i - 1], stops[i]) for i in range(1, len(stops))]
    
    def pred(self, given, key=None, givencov=None, fromdata=None, raw=False, keepcorr=None):
        """
        
        Compute the posterior.
        
        The posterior can be computed either for all points or for a subset,
        and either directly from data or from a posterior obtained with a fit.
        The latter case is for when the Gaussian process was used in a fit with
        other parameters.
        
        The output is a collection of gvars, either an array or a dictionary
        of arrays. They are properly correlated with gvars returned by
        `prior` and with the input data/fit.
        
        The input is a dictionary of arrays, `given`, with keys corresponding
        to the keys in the GP as added by `addx` or `addtransf`.
        
        Parameters
        ----------
        given : dictionary of arrays
            The data or fit result for some/all of the points in the GP.
            The arrays can contain either gvars or normal numbers, the latter
            being equivalent to zero-uncertainty gvars.
        key : None, key or list of keys, optional
            If None, compute the posterior for all points in the GP (also those
            used in `given`). Otherwise only those specified by key.
        givencov : dictionary of arrays, optional
            Covariance matrix of `given`. If not specified, the covariance
            is extracted from `given` with `gvar.evalcov(given)`.
        fromdata : bool
            Mandatory. Specify if the contents of `given` are data or already
            a posterior.
        raw : bool, optional
            If True, instead of returning a collection of gvars, return
            the mean and the covariance. When the mean is a dictionary, the
            covariance is a dictionary whose keys are pairs of keys of the
            mean (the same format used by `gvar.evalcov`). Default False.
        keepcorr : bool, optional
            If True (default), the returned gvars are correlated with the
            prior and the data/fit. If False, they have the correct covariance
            between themselves, but are independent from all other preexisting
            gvars.
        
        Returns
        -------
        If raw=False (default):
        
        posterior : array or dictionary of arrays
            A collections of gvars representing the posterior.
        
        If raw=True:
        
        pmean : array or dictionary of arrays
            The mean of the posterior. Equivalent to `gvar.mean(posterior)`.
        pcov : 2D array or dictionary of 2D arrays
            The covariance matrix of the posterior. If `pmean` is a dictionary,
            the keys of `pcov` are pairs of keys of `pmean`. Equivalent to
            `gvar.evalcov(posterior)`.
        
        """
        
        if fromdata is None:
            raise ValueError('you must specify if `given` is data or fit result')
        fromdata = bool(fromdata)
        raw = bool(raw)
        if keepcorr is None:
            keepcorr = not raw
        if keepcorr and raw:
            raise ValueError('both keepcorr=True and raw=True')
        
        strip = False
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = [key]
            strip = True
        outslices = self._slices(outkeys)
        
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = _concatenate_noop(ylist)
        
        Kxxs = self._assemblecovblocks(inkeys, outkeys)
        
        if ycovblocks is not None:
            ycov = _block_matrix(ycovblocks)
        elif (fromdata or raw or not keepcorr) and y.dtype == object:
            ycov = gvar.evalcov(gvar.gvar(y))
            # TODO use evalcov_blocks
        else:
            ycov = None
            
        if raw or not keepcorr or self._checkfinite:
            ymean = gvar.mean(y)
        if self._checkfinite and not np.all(np.isfinite(ymean)):
            raise ValueError('mean of `given` is not finite')
        if ycov is not None:
            if self._checkfinite and not np.all(np.isfinite(ycov)):
                raise ValueError('covariance matrix of `given` is not finite')
            if self._checksym and not np.allclose(ycov, ycov.T):
                raise ValueError('covariance matrix of `given` is not symmetric')
        
        if raw or not keepcorr:
            
            Kxsxs = self._assemblecovblocks(outkeys)
            
            if fromdata:
                solver = self._solver(inkeys, ycov)
            else:
                solver = self._solver(inkeys)

            mean = solver.quad(Kxxs, ymean)
            cov = Kxsxs - solver.quad(Kxxs)
            
            if not fromdata:
                # complete formula: cov = Kxsxs - A.T @ (Kxx - ycov) @ A
                if ycov is None:
                    pass
                else:
                    A = solver.solve(Kxxs)
                    cov = cov + A.T @ ycov @ A
            
        else: # (keepcorr and not raw)        
            yplist = [np.reshape(self._prior(key), -1) for key in inkeys]
            ysplist = [np.reshape(self._prior(key), -1) for key in outkeys]
            yp = _concatenate_noop(yplist)
            ysp = _concatenate_noop(ysplist)
            
            mat = ycov if fromdata else None
            flatout = ysp + self._solver(inkeys, mat).quad(Kxxs, y - yp)
        
        if raw and not strip:
            meandict = {
                key: mean[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            }
            
            covdict = {
                (row, col):
                cov[rowslice, colslice].reshape(self._elements[row].shape + self._elements[col].shape)
                for row, rowslice in zip(outkeys, outslices)
                for col, colslice in zip(outkeys, outslices)
            }
            
            return meandict, covdict
            
        elif raw:
            assert len(outkeys) == 1
            mean = mean.reshape(self._elements[outkeys[0]].shape)
            cov = cov.reshape(2 * self._elements[outkeys[0]].shape)
            return mean, cov
        
        elif not keepcorr:
            flatout = gvar.gvar(mean, cov, fast=True)
        
        if not strip:
            return gvar.BufferDict({
                key: flatout[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            })
            # TODO A Bufferdict can be initialized directly with the 1D buffer
        else:
            assert len(outkeys) == 1
            return flatout.reshape(self._elements[outkeys[0]].shape)
        
    def predfromfit(self, *args, **kw):
        """
        Like `pred` with `fromdata=False`.
        """
        return self.pred(*args, fromdata=False, **kw)
    
    def predfromdata(self, *args, **kw):
        """
        Like `pred` with `fromdata=True`.
        """
        return self.pred(*args, fromdata=True, **kw)

    def marginal_likelihood(self, given, givencov=None, separate=False):
        """
        
        Compute the logarithm of the probability of the data.
        
        The probability is computed under the Gaussian prior and Gaussian error
        model. It is also called marginal likelihood or bayes factor. If `y` is
        the data and `g` is the Gaussian process, this is
        
        .. math::
            \\log \\int p(y|g) p(g) \\mathrm{d} g.
        
        Unlike `pred()`, you can't compute this with a fit result instead of
        data. If you used the Gaussian process as latent variable in a fit,
        use the whole fit to compute the marginal likelihood. E.g. `lsqfit`
        always computes the logGBF (it's the same thing).
        
        The input is an array or dictionary of arrays, `given`. The contents of
        `given` represent the input data.
                
        Parameters
        ----------
        given : dictionary of arrays
            The data for some/all of the points in the GP. The arrays can
            contain either gvars or normal numbers, the latter being
            equivalent to zero-uncertainty gvars.
        givencov : dictionary of arrays, optional
            Covariance matrix of `given`. If not specified, the covariance
            is extracted from `given` with `gvar.evalcov(given)`.
        separate : bool
            If True, return separately the logdet term and the residuals term.
            Default False.
        
        Returns
        -------
        If ``separate == False`` (default):
        
        logp : scalar
            The logarithm of the marginal likelihood.
        
        If ``separate == True``:
            
        logdet : scalar
            The term of the marginal likelihood containing the logarithm of
            the determinant of the prior covariance matrix, without the -1/2
            factor.
        residuals : array or dictionary of arrays
            A vector whose 2-norm multiplied by -1/2 gives the other term of
            the marginal likelihood.
        """        
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = _concatenate_noop(ylist)
        
        if ycovblocks is not None:
            ycov = _block_matrix(ycovblocks)
            if y.dtype == object:
                warnings.warn(f'covariance matrix may have been specified both explicitly and with gvars; the explicit one will be used')
            ymean = gvar.mean(y)
        elif y.dtype == object:
            gvary = gvar.gvar(y)
            ycov = gvar.evalcov(gvary)
            ymean = gvar.mean(gvary)
        else:
            ycov = None
            ymean = y
        
        if self._checkfinite and not np.all(np.isfinite(ymean)):
            raise ValueError('mean of `given` is not finite')
        if ycov is not None:
            if self._checkfinite and not np.all(np.isfinite(ycov)):
                raise ValueError('covariance matrix of `given` is not finite')
            if self._checksym and not np.allclose(ycov, ycov.T):
                raise ValueError('covariance matrix of `given` is not symmetric')
        
        decomp = self._solver(inkeys, ycov)
        logdet = decomp.logdet() + len(y) * np.log(2 * np.pi)
        if separate:
            residuals = decomp.decorrelate(ymean)
            return logdet, residuals
        else:
            return -1/2 * (decomp.quad(ymean) + logdet)



# lsqfitgp/_array.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import builtins

from autograd import numpy as np
from autograd.builtins import isinstance

__all__ = [
    'StructuredArray',
    'broadcast_shapes',
    'broadcast',
    'broadcast_to',
    'broadcast_arrays',
    'asarray'
]

# TODO use the __array_function__ mechanism instead of exposing custom
# functions. Requires numpy >= 1.17. Is it still experimental, or will it stay
# as it is?

def _readonlyview(x):
    if not builtins.isinstance(x, (StructuredArray, np.numpy_boxes.ArrayBox)):
        x = x.view()
        x.flags['WRITEABLE'] = False
    return x

def _wrapifstructured(x):
    if x.dtype.names is None:
        return x
    else:
        return StructuredArray(x)

def _broadcast_shapes_2(s1, s2):
    assert isinstance(s1, tuple)
    assert isinstance(s2, tuple)
    if len(s1) < len(s2):
        s1 = (len(s2) - len(s1)) * (1,) + s1
    elif len(s2) < len(s1):
        s2 = (len(s1) - len(s2)) * (1,) + s2
    out = ()
    for a, b in zip(s1, s2):
        if a == b:
            out += (a,)
        elif a == 1 or b == 1:
            out += (a * b,)
        else:
            raise ValueError('can not broadcast shape {} with {}'.format(s1, s2))
    return out

def broadcast_shapes(shapes):
    """
    Return the broadcasted shape from a list of shapes.
    """
    out = ()
    for shape in shapes:
        try:
            out = _broadcast_shapes_2(out, shape)
        except ValueError:
            msg = 'can not broadcast shapes '
            msg += ', '.join(str(s) for s in shapes)
            raise ValueError(msg)
    return out

class broadcast:
    """
    Version of np.broadcast that works with StructuredArray.
    """
    
    def __init__(self, *arrays):
        shapes = [a.shape for a in arrays]
        self.shape = broadcast_shapes(shapes)

def broadcast_to(x, shape, **kw):
    """
    Version of np.broadcast_to that works with StructuredArray.
    """
    if isinstance(x, StructuredArray):
        return x.broadcast_to(shape, **kw)
    else:
        return np.broadcast_to(x, shape, **kw)

def broadcast_arrays(*arrays, **kw):
    """
    Version of np.broadcast_arrays that works with StructuredArray.
    """
    shapes = [a.shape for a in arrays]
    shape = broadcast_shapes(shapes)
    return tuple(broadcast_to(a, shape, **kw) for a in arrays)

def asarray(x, **kw):
    """
    Version of np.asarray that works with StructuredArray.
    """
    if isinstance(x, StructuredArray):
        return x
    else:
        return np.asarray(x)

class StructuredArray:
    """
    Autograd-friendly imitation of a numpy structured array.
    
    It behaves like a read-only numpy structured array, with the exception that
    you can set a whole field/subfield.
    
    Examples
    --------
    >>> a = np.empty(3, dtype=[('f', float), ('g', float)])
    >>> a = StructuredArray(a)
    >>> a['f'] = np.arange(3) # this is allowed
    >>> a[1] = (0.3, 0.4) # this raises an error
        
    Parameters
    ----------
    array : numpy array, StructuredArray
        A structured array. An array qualifies as structured if
        ``array.dtype.names is not None``.
    
    Notes
    -----
    The StructuredArray is a readonly view on the input array. When you
    change the content of a field of the StructuredArray, however, the
    reference to the original array for that field is lost.
    
    """
    
    @classmethod
    def _fromarrayanddict(cls, x, d, readonly=False):
        """
        Create a new StructuredArray, copying the dtype member from `x` and
        using `d` as `_dict` member (no copy). The shape and size are inferred
        from the contents of `d`.
        """
        out = super().__new__(cls)
        out.dtype = x.dtype
        out._dict = d
        f0 = x.dtype.names[0]
        a0 = d[f0]
        subshape = x.dtype.fields[f0][0].shape
        out.shape = a0.shape[:len(a0.shape) - len(subshape)]
        out.size = np.prod(out.shape)
        out._readonly = readonly
        return out
    
    def __new__(cls, array):
        assert isinstance(array, (np.ndarray, cls))
        assert array.dtype.names is not None
        d = {
            name: _readonlyview(_wrapifstructured(array[name]))
            for name in array.dtype.names
        }
        return cls._fromarrayanddict(array, d)
    
    def __getitem__(self, key):
        if isinstance(key, str):
            return self._dict[key]
        elif isinstance(key, list) and all(isinstance(k, str) for k in key):
            d = {
                name: self._dict[name]
                for name in key
            }
            return type(self)._fromarrayanddict(self, d)
        else:
            d = {
                name: x[
                    (key if isinstance(key, tuple) else (key,))
                    + (slice(None),) * len(self.dtype.fields[name][0].shape)
                ]
                for name, x in self._dict.items()
            }
            return type(self)._fromarrayanddict(self, d, readonly=True)
    
    def __setitem__(self, key, val):
        if self._readonly:
            msg = 'This StructuredArray is read-only, maybe you did '
            msg += '"array[index][label] = ..." instead of '
            msg += '"array[label] = np.array([...])"?'
            raise ValueError(msg)
        assert key in self.dtype.names
        assert isinstance(val, (np.ndarray, StructuredArray))
        prev = self._dict[key]
        # TODO support casting and broadcasting
        assert prev.dtype == val.dtype
        assert prev.shape == val.shape
        self._dict[key] = _readonlyview(val)
    
    def reshape(self, *shape):
        """
        Reshape the array without changing its contents. See np.ndarray.reshape.
        """
        if len(shape) == 1 and isinstance(shape[0], tuple):
            shape = shape[0]
        d = {
            name: x.reshape(shape + self.dtype.fields[name][0].shape)
            for name, x in self._dict.items()
        }
        return type(self)._fromarrayanddict(self, d)
    
    def broadcast_to(self, shape, **kw):
        """
        Return a view of the array broadcasted to another shape. See
        np.broadcast_to.
        """
        _broadcast_shapes_2(self.shape, shape) # raises if not broadcastable
        d = {
            name: broadcast_to(x, shape + self.dtype.fields[name][0].shape, **kw)
            for name, x in self._dict.items()
        }
        return type(self)._fromarrayanddict(self, d)



# lsqfitgp/_fit.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import warnings

import gvar
import autograd
from autograd import numpy as np
from scipy import optimize

from . import _GP
from . import _linalg

__all__ = [
    'empbayes_fit'
]

@autograd.wrap_util.unary_to_nary
def jac(fun, x): # pragma: no cover
    """Like autograd.jacobian but with forward mode, currently not used"""
    jvp = autograd.core.make_jvp(fun, x)
    ans = fun(x)
    vs = autograd.extend.vspace(x)
    grads = map(lambda b: jvp(b)[1], vs.standard_basis())
    return np.reshape(np.stack(grads, axis=-1), ans.shape + vs.shape)

def _asarrayorbufferdict(x):
    if hasattr(x, 'keys'):
        return gvar.BufferDict(x)
    else:
        return np.array(x, copy=False)

def _flat(x):
    if isinstance(x, np.ndarray):
        return x.reshape(-1)
    elif isinstance(x, gvar.BufferDict):
        return x.buf

def _unflat(x, original):
    if isinstance(original, np.ndarray):
        out = x.reshape(original.shape)
        return out if out.shape else out.item()
    elif isinstance(original, gvar.BufferDict):
        return gvar.BufferDict(original, buf=x)

class empbayes_fit:

    def __init__(self, hyperprior, gpfactory, data, raises=True, minkw={}, gpfactorykw={}):
        """
    
        Empirical bayes fit.
    
        Maximizes the marginal likelihood of the data with a Gaussian process
        model that depends on hyperparameters.
    
        Parameters
        ----------
        hyperprior : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the prior for the
            hyperparameters.
        gpfactory : callable
            A function with signature gpfactory(hyperparams) -> GP object. The
            argument `hyperparams` has the same structure of the empbayes_fit
            argument `hyperprior`. gpfactory must be autograd-friendly, i.e.,
            either use autograd.numpy, autograd.scipy, lsqfitgp.numpy,
            lsqfitgp.scipy or gvar instead of plain numpy/scipy.
        data : dict or callable
            Dictionary of data that is passed to `GP.marginal_likelihood` on
            the GP object returned by `gpfactory`. If a callable, it is called
            with the same arguments of `gpfactory` and must return either a
            dictionary or a pair of dictionaries where the second dictionary is
            passed as `givencov` argument to `GP.marginal_likelihood`.
        raises : bool, optional
            If True (default), raise an error when the minimization fails.
            Otherwise, use the last point of the minimization as result.
        minkw : dict, optional
            Keyword arguments passed to `scipy.optimize.minimize`.
        gpfactorykw : dict, optional
            Keyword arguments passed to `gpfactory`, and also to `data` if it
            is a callable.
    
        Attributes
        ----------
        p : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the hyperparameters that
            maximize the marginal likelihood. The covariance matrix is computed
            as the inverse of the hessian of the marginal likelihood. These
            gvars do not track correlations with the hyperprior or the data.
        minresult : scipy.optimize.OptimizeResult
            The result object returned by `scipy.optimize.minimize`.

        Raises
        ------
        RuntimeError
            The minimization failed and `raises` is True.
    
        """
        assert callable(gpfactory)
    
        # TODO would it be meaningful to add correlation of the fit result with
        # the data and hyperprior?
    
        # TODO add the second order correction. It probably requires more than
        # the gradient and inv_hess, but maybe by getting a little help from
        # marginal_likelihood I can use the least-squares optimized second order
        # correction on the residuals term and invent something for the logdet
        # term.
    
        # TODO it raises very often with "Desired error not necessarily
        # achieved due to precision loss.". Change the default arguments of
        # minimize to make this less frequent, but only after implement the
        # quad-specific derivatives since maybe they will fix this. Another
        # thing that may matter is the vjp of the logdet, which computes a
        # matrix inverse. I could do the following: make an internal version of
        # marginal_likelihood that returns separately the residuals and the
        # logdet term, and do a backward derivative on the residuals and a
        # forward on the logdet. Other option: try another optimization
        # algorithm.
        
        # TODO I don't know really how much the inverse hessian estimated by
        # BFGS is accurate. Investigate computing the hessian with autograd or
        # using Gauss-Newton on the residuals and autograd on the logdet.
        
        # TODO read the minkw options to adapt the usage of the jacobian: if
        # jac is specified or if the method does not require derivatives do not
        # take derivatives with autograd.
        
        # TODO compute the logGBF for the whole fit.
    
        hyperprior = _asarrayorbufferdict(hyperprior)
        flathp = _flat(hyperprior)
        hpmean = gvar.mean(flathp)
        hpcov = gvar.evalcov(flathp) # TODO use evalcov_blocks (low priority)
        hpdec = _linalg.EigCutFullRank(hpcov)
        
        if callable(data):
            cachedargs = None
        elif _flat(_asarrayorbufferdict(data)).dtype == object:
            data = gvar.gvar(data)
            datamean = gvar.mean(data)
            datacov = gvar.evalcov(data)
            cachedargs = (datamean, datacov)
        else:
            cachedargs = (data,)
    
        def fun(p):
            priorchi2 = hpdec.quad(p - hpmean)

            hp = _unflat(p, hyperprior)
            gp = gpfactory(hp, **gpfactorykw)
            assert isinstance(gp, _GP.GP)
            
            if cachedargs:
                args = cachedargs
            else:
                args = data(hp, **gpfactorykw)
                if not isinstance(args, tuple):
                    args = (args,)
            ml = gp.marginal_likelihood(*args)
            
            return -ml + 1/2 * priorchi2
    
        result = optimize.minimize(autograd.value_and_grad(fun), hpmean, jac=True, **minkw)
        
        if not result.success:
            msg = 'minimization failed: {}'.format(result.message)
            if raises:
                raise RuntimeError(msg)
            else:
                warnings.warn(msg)
        
        uresult = gvar.gvar(result.x, result.hess_inv)
        
        self.p = _unflat(uresult, hyperprior)
        self.minresult = result



# lsqfitgp/_linalg.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

Decompositions of positive definite matrices. A decomposition object is
initialized with a matrix and then can solve linear systems for that matrix.
These classes never check for infs/nans in the matrices.

Classes
-------
Decomposition
    Abstract base class.
DecompAutograd
    Abstract subclass that adds autograd support.
Diag
    Diagonalization.
EigCutFullRank
    Diagonalization rounding up small (or negative) eigenvalues.
EigCutLowRank
    Diagonalization removing small (or negative) eigenvalues.
SVDCutFullRank
    Diagonalization rounding up small eigenvalues, keeping their sign.
SVDCutLowRank
    Diagonalization removing small eigenvalues.
ReduceRank
    Partial diagonalization with higher eigenvalues only.
Chol
    Cholesky decomposition.
CholReg
    Abstract base class for regularized Cholesky decomposition.
CholMaxEig
    Cholesky regularized using the maximum eigenvalue.
CholGersh
    Cholesky regularized using an estimate of the maximum eigenvalue.
CholToeplitz
    Cholesky decomposition of a Toeplitz matrix regularized using an estimate
    of the maximum eigenvalue.
BlockDecomp
    Decompose a 2x2 block matrix.
BlockDiagDecomp
    Decompose a 2x2 block diagonal matrix.

"""

import abc
import functools

import autograd
from autograd import numpy as np
from scipy import linalg
from scipy import sparse

from . import _toeplitz_linalg

# TODO optimize the matrix multiplication with gvars. Use these gvar internals:
# gvar.svec(int size)
# gvar.svec._assign(float[] values, int[] indices)
# gvar.GVar(float mean, svec derivs, smat cov)
# it may require cython to be fast since it's not vectorized

# TODO investigate using the QR decomposition. It's twice as fast as
# diagonalization, it does not require positivity, it allows tweaking the
# generalized eigenvalues (the diagonal of the triangular factor). Does not
# support the methods correlate and decorrelate. Low-rank updates already
# provided by scipy.

def noautograd(x):
    """
    Unpack an autograd numpy array.
    """
    if isinstance(x, np.numpy_boxes.ArrayBox):
        return noautograd(x._value)
    else:
        return x

def asinexact(dtype):
    """
    Return dtype if it is inexact, else float64.
    """
    if np.issubdtype(dtype, np.inexact):
        return dtype
    else:
        return np.float64

class Decomposition(metaclass=abc.ABCMeta):
    """
    
    Abstract base class for positive definite symmetric matrices decomposition.
    
    Methods
    -------
    solve
    quad
    logdet
    correlate
    decorrelate
    inv
    
    Properties
    ----------
    n
    
    """
    
    @abc.abstractmethod
    def __init__(self, K): # pragma: no cover
        """
        Decompose matrix K.
        """
        pass
        
    @abc.abstractmethod
    def solve(self, b): # pragma: no cover
        """
        Solve the linear system K @ x = b. `b` can be an array of gvars.
        """
        pass
    
    def quad(self, b, c=None):
        """
        Compute the quadratic form b.T @ inv(K) @ b if c is not specified, else
        b.T @ inv(K) @ c. `b` and `c` can be arrays of gvars.
        """
        if c is None:
            c = b
        return b.T @ self.solve(c)
    
    @abc.abstractmethod
    def logdet(self): # pragma: no cover
        """
        Compute log(det(K)).
        """
        pass
    
    @abc.abstractmethod
    def correlate(self, b): # pragma: no cover
        """
        Compute A @ b where K = A @ A.T. If b represents iid variables with
        unitary variance, A @ b has covariance matrix K.
        """
        pass
    
    @abc.abstractmethod
    def decorrelate(self, b): # pragma: no cover
        """
        Solve A @ x = b, where K = A @ A.T. If b represents variables with
        covariance matrix K, x has identity covariance.
        """
        pass
    
    def inv(self):
        """
        Invert the matrix.
        """
        
        # TODO currently no subclass overrides this, some should probably do.
        
        return self.quad(np.eye(self.n))
    
    @property
    def n(self):
        """
        Return n where the decomposed matrix is n x n.
        """
        return len(self._K)

class DecompAutograd(Decomposition):
    """
    Abstract subclass adding autograd support to subclasses of Decomposition.
    Moreover, it fixes the convention of tensor multiplication by passing
    the concrete class methods only 2d matrices and then reshaping back the
    result.
    """
        
    def __init_subclass__(cls, **kw):
        
        # For __init__ I can't use an _autograd flag like below to avoid double
        # wrapping because the wrapper is called as super().__init__ in
        # subclasses, so I assign self._K *after* calling old__init__.
        
        old__init__ = cls.__init__
        
        @functools.wraps(old__init__)
        def __init__(self, K, **kw):
            old__init__(self, noautograd(K), **kw)
            self._K = K
        
        cls.__init__ = __init__
        
        for name in 'solve', 'quad', 'logdet':
            meth = getattr(cls, name)
            if not hasattr(meth, '_autograd'):
                # print(f'defining {cls.__name__}.{name}')
                newmeth = getattr(DecompAutograd, '_make_' + name)(meth)
                newmeth = functools.wraps(meth)(newmeth)
                if not hasattr(newmeth, '_autograd'):
                    newmeth._autograd = True
                setattr(cls, name, newmeth)
        
        super().__init_subclass__(**kw)

    @staticmethod
    def _make_solve(oldsolve):
        
        @autograd.extend.primitive
        def solve_autograd(self, K, b):
            b2d = b.reshape(b.shape[0], -1)
            return oldsolve(self, b2d).reshape(b.shape)
        
        def solve_vjp_K(ans, self, K, b):
            assert ans.shape == b.shape
            assert b.shape[0] == K.shape[0] == K.shape[1]
            def vjp(g):
                assert g.shape[-len(b.shape):] == b.shape
                g = np.moveaxis(g, -len(b.shape), 0)
                A = solve_autograd(self, K, g)
                B = np.moveaxis(ans, 0, -1)
                AB = np.tensordot(A, B, len(b.shape) - 1)
                AB = np.moveaxis(AB, 0, -2)
                assert AB.shape == g.shape[:-len(b.shape)] + K.shape
                return -AB
            return vjp
        
        def solve_vjp_b(ans, self, K, b):
            assert ans.shape == b.shape
            assert b.shape[0] == K.shape[0] == K.shape[1]
            def vjp(g):
                assert g.shape[-len(b.shape):] == b.shape
                g = np.moveaxis(g, -len(b.shape), 0)
                gj = solve_autograd(self, K, g)
                gj = np.moveaxis(gj, 0, -len(b.shape))
                assert gj.shape == g.shape
                return gj
            return vjp
        
        autograd.extend.defvjp(
            solve_autograd,
            solve_vjp_K,
            solve_vjp_b,
            argnums=[1, 2]
        )
        
        def solve_jvp_K(g, ans, self, K, b):
            assert ans.shape == b.shape
            assert ans.shape[0] == K.shape[0] == K.shape[1]
            assert g.shape[:2] == K.shape
            invKg = solve_autograd(self, K, g)
            ans2D = ans.reshape(ans.shape[0], -1)
            jvp = -np.tensordot(invKg, ans2D, axes=(1, 0))
            jvp = np.moveaxis(jvp, -1, 1)
            return jvp.reshape(ans.shape + g.shape[2:])
        
        def solve_jvp_b(g, ans, self, K, b):
            assert ans.shape == b.shape
            assert b.shape[0] == K.shape[0] == K.shape[1]
            assert g.shape[:len(b.shape)] == b.shape
            return solve_autograd(self, K, g)
        
        autograd.extend.defjvp(
            solve_autograd,
            solve_jvp_K,
            solve_jvp_b,
            argnums=[1, 2]
        )
        
        def solve(self, b):
            return solve_autograd(self, self._K, b)
        
        # solve_autograd is used by logdet_vjp, so I store it here in case
        # logdet but not solve needs wrapping in a subclass
        solve._autograd = solve_autograd
        
        return solve
    
    @staticmethod
    def _make_quad(oldquad):
        
        @autograd.extend.primitive
        def quad_autograd(self, K, b, c):
            b2d = b.reshape(b.shape[0], -1)
            if c is None:
                outshape = b.shape[1:][::-1] + b.shape[1:]
                return oldquad(self, b2d).reshape(outshape)
            else:
                c2d = c.reshape(c.shape[0], -1)
                outshape = b.shape[1:][::-1] + c.shape[1:]
                return oldquad(self, b2d, c2d).reshape(outshape)

        def quad_vjp_K(ans, self, K, b, c):
            assert b.shape[0] == K.shape[0] == K.shape[1]
            bshape = b.shape[1:]
            if c is None:
                cshape = bshape
            else:
                assert c.shape[0] == b.shape[0]
                cshape = c.shape[1:]
            assert ans.shape == tuple(reversed(bshape)) + cshape

            def vjp(g):
                assert g.shape[len(g.shape) - len(ans.shape):] == ans.shape
                
                invKb = self.solve._autograd(self, K, b)
                if c is None:
                    invKc = invKb
                else:
                    invKc = self.solve._autograd(self, K, c)

                axes = 2 * (tuple(range(-len(cshape), 0)),)
                ginvKc = np.tensordot(g, invKc, axes)

                axes = 2 * (tuple(range(-len(bshape) - 1, -1)),)
                ginvKcb = np.tensordot(ginvKc, invKb.T, axes)

                assert ginvKcb.shape == g.shape[:len(g.shape) - len(ans.shape)] + K.shape
                return -ginvKcb
            return vjp

        def quad_vjp_b(ans, self, K, b, c):
            assert b.shape[0] == K.shape[0] == K.shape[1]
            bshape = b.shape[1:]
            if c is None:
                cshape = bshape
            else:
                assert c.shape[0] == b.shape[0]
                cshape = c.shape[1:]
            assert ans.shape == tuple(reversed(bshape)) + cshape

            if c is None:
                def vjp(g):
                    glen = len(g.shape) - len(ans.shape)
                    assert g.shape[glen:] == ans.shape
            
                    invKb = self.solve._autograd(self, K, b)

                    axes = 2 * (range(-len(bshape), 0),)
                    gj1 = np.tensordot(g, invKb, axes)

                    axes = tuple(range(glen))
                    axes += tuple(reversed(range(glen, len(gj1.shape))))
                    gj1 = np.transpose(gj1, axes)
                    assert gj1.shape == g.shape[:glen] + b.shape
                    
                    axes = (
                        tuple(range(-len(bshape) - 1, -2 * len(bshape) - 1, -1)),
                        tuple(range(-len(bshape), 0))
                    )
                    gj2 = np.tensordot(g, invKb, axes)

                    gj2 = np.moveaxis(gj2, -1, -len(bshape) - 1)
                    assert gj2.shape == gj1.shape

                    return gj1 + gj2
            else:
                def vjp(g):
                    glen = len(g.shape) - len(ans.shape)
                    assert g.shape[glen:] == ans.shape
            
                    invKc = self.solve._autograd(self, K, c)

                    axes = 2 * (range(-len(cshape), 0),)
                    gj = np.tensordot(g, invKc, axes)

                    axes = tuple(range(glen))
                    axes += tuple(reversed(range(glen, len(gj.shape))))
                    gj = np.transpose(gj, axes)
                    assert gj.shape == g.shape[:glen] + b.shape
                    return gj
            
            return vjp

        def quad_vjp_c(ans, self, K, b, c):
            assert b.shape[0] == K.shape[0] == K.shape[1]
            assert c.shape[0] == b.shape[0]
            bshape = b.shape[1:]
            cshape = c.shape[1:]
            assert ans.shape == tuple(reversed(bshape)) + cshape

            def vjp(g):
                assert g.shape[len(g.shape) - len(ans.shape):] == ans.shape
            
                invKb = self.solve._autograd(self, K, b)

                axes = (
                    tuple(range(-len(cshape) - 1, -len(cshape) - len(bshape) - 1, -1)),
                    tuple(range(-len(bshape), 0))
                )
                gj = np.tensordot(g, invKb, axes)

                gj = np.moveaxis(gj, -1, -len(cshape) - 1)
                assert gj.shape == g.shape[:len(g.shape) - len(ans.shape)] + c.shape
                return gj
            
            return vjp

        autograd.extend.defvjp(
            quad_autograd,
            quad_vjp_K,
            quad_vjp_b,
            quad_vjp_c,
            argnums=[1, 2, 3]
        )
        
        def quad_jvp_K(g, ans, self, K, b, c):
            if c is None:
                c = b
            bshape = b.shape[1:]
            cshape = c.shape[1:]
            assert ans.shape == tuple(reversed(bshape)) + cshape
            assert g.shape[:2] == K.shape
            g3d = g.reshape(K.shape + (-1,))
            ginvKc = quad_autograd(self, K, g3d, c)
            ginvKc = np.moveaxis(ginvKc, 0, -1)
            ginvKc = ginvKc.reshape(c.shape + g.shape[2:])
            jvp = -quad_autograd(self, K, b, ginvKc)
            assert jvp.shape == ans.shape + g.shape[2:]
            return jvp
        
        def quad_jvp_b(g, ans, self, K, b, c):
            assert g.shape[:len(b.shape)] == b.shape
            g1d = g.reshape(b.shape + (-1,))
            if c is None:
                jvp = quad_autograd(self, K, b, g1d)
                jvpT = np.moveaxis(jvp.T, 0, -1)
                jvp = jvp + jvpT
            else:
                jvp = quad_autograd(self, K, g1d, c)
                jvp = np.moveaxis(jvp, 0, -1)
            return jvp.reshape(ans.shape + g.shape[len(b.shape):])
        
        def quad_jvp_c(g, ans, self, K, b, c):
            assert g.shape[:len(c.shape)] == c.shape
            jvp = quad_autograd(self, K, b, g)
            assert jvp.shape == ans.shape + g.shape[len(c.shape):]
            return jvp
        
        autograd.extend.defjvp(
            quad_autograd,
            quad_jvp_K,
            quad_jvp_b,
            quad_jvp_c,
            argnums=[1, 2, 3]
        )

        def quad(self, b, c=None):
            return quad_autograd(self, self._K, b, c)
        quad._autograd = quad_autograd

        return quad
    
    @staticmethod
    def _make_logdet(oldlogdet):
        
        @autograd.extend.primitive
        def logdet_autograd(self, K):
            return oldlogdet(self)
        
        def logdet_vjp(ans, self, K):
            assert ans.shape == ()
            assert K.shape[0] == K.shape[1]
            def vjp(g):
                invK = self.quad._autograd(self, K, np.eye(len(K)), None)
                return g[..., None, None] * invK
            return vjp
        
        autograd.extend.defvjp(
            logdet_autograd,
            logdet_vjp,
            argnums=[1]
        )
        
        def logdet_jvp(g, ans, self, K):
            assert ans.shape == ()
            assert K.shape[0] == K.shape[1]
            assert g.shape[:2] == K.shape
            return np.trace(self.solve._autograd(self, K, g))
        
        autograd.extend.defjvp(
            logdet_autograd,
            logdet_jvp,
            argnums=[1]
        )
        
        def logdet(self):
            return logdet_autograd(self, self._K)
        
        return logdet
    
class Diag(DecompAutograd):
    """
    Diagonalization.
    """
    
    def __init__(self, K):
        self._w, self._V = linalg.eigh(K, check_finite=False)
    
    def solve(self, b):
        return (self._V / self._w) @ (self._V.T @ b)
    
    def quad(self, b, c=None):
        VTb = self._V.T @ b
        if c is None:
            VTc = VTb
        else:
            VTc = self._V.T @ c
        return (VTb.T / self._w) @ VTc
    
    def logdet(self):
        return np.sum(np.log(self._w))
    
    def correlate(self, b):
        return (self._V * np.sqrt(self._w)) @ b
    
    def decorrelate(self, b):
        return (self._V / np.sqrt(self._w)).T @ b
    
    def _eps(self, eps):
        w = self._w
        if eps is None:
            eps = len(w) * np.finfo(asinexact(w.dtype)).eps
        assert np.isscalar(eps) and 0 <= eps < 1
        return eps * np.max(w)

class EigCutFullRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` are set to `eps`, where `eps` is
    relative to the largest eigenvalue.
    """
    
    def __init__(self, K, eps=None):
        super().__init__(K)
        eps = self._eps(eps)
        self._w[self._w < eps] = eps
            
class EigCutLowRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` are removed, where `eps` is
    relative to the largest eigenvalue.
    """
    
    def __init__(self, K, eps=None):
        super().__init__(K)
        eps = self._eps(eps)
        subset = slice(np.sum(self._w < eps), None) # w is sorted ascending
        self._w = self._w[subset]
        self._V = self._V[:, subset]
        
class SVDCutFullRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` in absolute value are set to
    `eps` with their sign, where `eps` is relative to the largest eigenvalue.
    """
    
    def __init__(self, K, eps=None):
        super().__init__(K)
        eps = self._eps(eps)
        cond = np.abs(self._w) < eps
        self._w[cond] = eps * np.sign(self._w[cond])
        
class SVDCutLowRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` in absolute value are removed,
    where `eps` is relative to the largest eigenvalue.
    """
    
    def __init__(self, K, eps=None):
        super().__init__(K)
        eps = self._eps(eps)
        subset = np.abs(self._w) >= eps
        self._w = self._w[subset]
        self._V = self._V[:, subset]

class ReduceRank(Diag):
    """
    Keep only the first `rank` higher eigenmodes.
    """
    
    def __init__(self, K, rank=1):
        assert isinstance(rank, (int, np.integer)) and rank >= 1
        self._w, self._V = sparse.linalg.eigsh(K, k=rank, which='LM')
    
    def correlate(self, b):
        return super().correlate(b[:len(self._w)])

def solve_triangular(a, b, lower=False):
    """
    Pure python implementation of scipy.linalg.solve_triangular for when
    a or b are object arrays. Differently from the scipy version, it
    satisfies tensordot(a, solve_triangular(a, b), 1) == b instead of
    a @ solve_triangular(a, b) == b. It makes a difference only if b is >2D.
    """
    # TODO maybe commit this to gvar.linalg
    # TODO can I raise a LinAlgError if a[i,i] is 0, and still return the
    # result and have it assigned to a variable using try...finally inside this
    # function? => Probably not
    x = np.copy(b)
    a = a.reshape(a.shape + (1,) * len(x.shape[1:]))
    if lower:
        x[0] /= a[0, 0]
        for i in range(1, len(x)):
            x[i:] -= x[i - 1] * a[i:, i - 1]
            x[i] /= a[i, i]
    else:
        x[-1] /= a[-1, -1]
        for i in range(len(x) - 1, 0, -1):
            x[:i] -= x[i] * a[:i, i]
            x[i - 1] /= a[i - 1, i - 1]
    return x

def solve_triangular_auto(a, b, lower=False):
    """Works with b both object and non-object array"""
    if b.dtype == object:
        return solve_triangular(a, b, lower=lower)
    else:
        return linalg.solve_triangular(a, b, lower=lower, check_finite=False)

class Chol(DecompAutograd):
    """
    Cholesky decomposition.
    """
    
    def __init__(self, K):
        self._L = linalg.cholesky(K, lower=True, check_finite=False)
    
    def solve(self, b):
        invLb = solve_triangular_auto(self._L, b, lower=True)
        return solve_triangular_auto(self._L.T, invLb, lower=False)
    
    def quad(self, b, c=None):
        invLb = solve_triangular_auto(self._L, b, lower=True)
        if c is None:
            invLc = invLb
        else:
            invLc = solve_triangular_auto(self._L, c, lower=True)
        return invLb.T @ invLc
    
    def logdet(self):
        return 2 * np.sum(np.log(np.diag(self._L)))
    
    def correlate(self, b):
        return self._L @ b
    
    def decorrelate(self, b):
        return solve_triangular_auto(self._L, b, lower=True)

    def _eps(self, eps, mat, maxeigv):
        if eps is None:
            eps = len(mat) * np.finfo(asinexact(mat.dtype)).eps
        assert np.isscalar(eps) and 0 <= eps < 1
        return eps * maxeigv

def _scale(a):
    """
    Compute a vector s of powers of 2 such that diag(a / outer(s, s)) ~ 1.
    """
    return np.exp2(np.rint(0.5 * np.log2(np.diag(a))))

class CholReg(Chol):
    """
    Cholesky decomposition correcting for roundoff. Abstract class.
    """
    
    def __init__(self, K, eps=None):
        s = _scale(K)
        J = K / s[:, None]
        J /= s[None, :]
        self._regularize(J, eps)
        super().__init__(J)
        self._L *= s[:, None]
    
    @abc.abstractmethod
    def _regularize(self, mat, eps): # pragma: no cover
        """Modify mat in-place to make it positive definite."""
        pass
    
def _gershgorin_eigval_bound(mat):
    """
    Upper bound on the largest magnitude eigenvalue of the matrix.
    """
    return np.max(np.sum(np.abs(mat), axis=1))

class CholGersh(CholReg):
    """
    Cholesky decomposition. The matrix is corrected for numerical roundoff
    by adding to the diagonal a small number relative to the maximum eigenvalue.
    `eps` multiplies this number. The maximum eigenvalue is estimated
    with the Gershgorin theorem.
    """
    
    def _regularize(self, mat, eps):
        maxeigv = _gershgorin_eigval_bound(mat)
        mat[np.diag_indices(len(mat))] += self._eps(eps, mat, maxeigv)

class CholToeplitz(Chol):
    """
    Cholesky decomposition of a Toeplitz matrix. Only the first row of the
    matrix is read.
    """
    
    def __init__(self, K, eps=None):
        t = K[0]
        m = _toeplitz_linalg.eigv_bound(t)
        eps = self._eps(eps, t, m)
        self._L = _toeplitz_linalg.cholesky(t, diageps=eps)

class BlockDecomp(Decomposition):
    """
    Decomposition of a 2x2 symmetric block matrix using decompositions of the
    diagonal blocks.
    
    Reference: Gaussian Processes for Machine Learning, A.3, p. 201.
    """
    
    # This class can be used only starting from a seed block and adding
    # other blocks one at a time. Would a divide et impera approach be useful
    # for my case?
    
    # TODO is it more efficient to make this a subclass of DecompAutograd?
    
    def __init__(self, P_decomp, S, Q, S_decomp_class):
        """
        The matrix to be decomposed is
        
            K = [[P,   Q]
                 [Q.T, S]]
        
        Parameters
        ----------
        P_decomp : Decomposition
            An instantiated decomposition of P.
        S, Q : matrices
            The other blocks.
        S_decomp_class : subclass of Decomposition
            A subclass of Decomposition used to decompose S - Q.T P^-1 Q.
        """
        self._Q = Q
        self._invP = P_decomp
        self._tildeS = S_decomp_class(S - P_decomp.quad(Q))
    
    def solve(self, b):
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        gQTinvPf = g - invP.quad(Q, f)
        y = tildeS.solve(gQTinvPf)
        x = invP.solve(f - tildeS.quad(Q.T, gQTinvPf))
        return np.concatenate([x, y])
    
    def quad(self, b, c=None):
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        QTinvPf = invP.quad(Q, f)
        if c is None:
            fTinvPQtildeSg = tildeS.quad(QTinvPf, g)
            gtildeSQTinvPf = fTinvPQtildeSg.T if fTinvPQtildeSg.shape else fTinvPQtildeSg
            return invP.quad(f) + tildeS.quad(QTinvPf) - fTinvPQtildeSg - gtildeSQTinvPf + tildeS.quad(g)
        else:
            h = c[:len(Q)]
            i = c[len(Q):]
            QTinvPh = invP.quad(Q, h)
            fTinvPQtildeSi = tildeS.quad(QTinvPf, i)
            gTtildeSQTinvPh = tildeS.quad(g, QTinvPh)
            return invP.quad(f, h) + tildeS.quad(QTinvPf, QTinvPh) - fTinvPQtildeSi - gTtildeSQTinvPh + tildeS.quad(g, i)
    
    def logdet(self):
        return self._invP.logdet() + self._tildeS.logdet()
    
    def correlate(self, b):
        # Block Cholesky decomposition:
        # K = [P    Q] = L L^T
        #     [Q^T  S]
        # L = [A   ]
        #     [B  C]
        # AA^T = P
        # AB^T = Q  ==>  B^T = A^-1Q
        # BB^T + CC^T = S  ==>  CC^T = S - Q^T P^-1 Q
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        x = invP.correlate(f)
        y = np.tensordot(invP.decorrelate(Q).T, f, 1) + tildeS.correlate(g)
        return np.concatenate([x, y])
    
    def decorrelate(self, b):
        # L^-1 = [   A^-1         ]
        #        [-C^-1BA^-1  C^-1]
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        x = invP.decorrelate(f)
        y = tildeS.decorrelate(g - invP.quad(Q, f))
        return np.concatenate([x, y])
        
    @property
    def n(self):
        return sum(self._Q.shape)

class BlockDiagDecomp(Decomposition):
    
    # TODO allow NxN instead of 2x2?
    
    def __init__(self, A_decomp, B_decomp):
        """
        
        Decomposition of a 2x2 block diagonal matrix.
        
        The matrix is
        
            [[A  0]
             [0  B]]
        
        Parameters
        ----------
        A_decomp, B_decomp: Decomposition
            Instantiated decompositions of A and B.
        
        """
        self._A = A_decomp
        self._B = B_decomp
    
    def solve(self, b):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        return np.concatenate([A.solve(f), B.solve(g)])
    
    def quad(self, b, c=None):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        if c is None:
            return A.quad(f) + B.quad(g)
        else:
            h = c[:An]
            i = c[An:]
            return A.quad(f, h) + B.quad(g, i)
    
    def logdet(self):
        return self._A.logdet() + self._B.logdet()
    
    def correlate(self, b):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        return np.concatenate([A.correlate(f), B.correlate(g)])
    
    def decorrelate(self, b):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        return np.concatenate([A.decorrelate(f), B.decorrelate(g)])

    @property
    def n(self):
        return self._A.n + self._B.n
